"""PyCorrFit data set:  Classes for FCS data evaluation."""
import hashlib
import warnings

import numpy as np

from . import models as mdls
from . import fit
from .trace import Trace


class Correlation(object):
    """ unifies correlation curve handling
    """

    def __init__(self, backgrounds=[], correlation=None, corr_type="AC",
                 filename=None, fit_algorithm="Lev-Mar",
                 fit_model=6000, fit_ival=(0, 0),
                 fit_weight_data=None, fit_weight_type="none",
                 normparm=None, title=None, traces=[], verbose=1):
        """
        Parameters
        ----------
        backgrounds: list of instances of Trace
            background traces
        correlation: ndarray of shape (N,2)
            correlation data (time [s], correlation)
        corr_type: str
            type of correlation, e.g. "AC", "AC1", "cc12"
        filename: str
            path to filename of correlation
        fit_algorithm: str
            valid fit algorithm identifier (Algorithms.keys())
        fit_ival:
            fitting interval of lag times in indices
        fit_model: instance of FitModel
            the model used for fitting
        fit_weight_data: any
            data for the certain fit_weight_type
        fit_weight_type: str
            Reserved keywords or user-defined strings:
             - "none" : no weights are used
             - "splineX" : compute weights from spline with X knots
                   and a spread of `fit_weight_data` bins.
             - "model function" : compute weights from difference
                   to model function
             - user-defined : other weights (e.g. previously computed
                   averages given in fit_weight_data)
        normparm: int
            identifier of normalization parameter
        title: str
            user-editable title of this correlation
        traces: list of instances of Trace
            traces of the current correlation
        verbose : int
            increment to increase verbosity
        """
        # must be created before setting properties
        self._backgrounds = []
        self._correlation = None
        self._fit_algorithm = None
        self._fit_model = None
        self._fit_parameters = None
        self._fit_parameters_range = None
        self._fit_parameters_variable = None
        self._fit_weight_memory = dict()
        self._lag_time = None
        self._model_memory = dict()
        self._traces = []
        self._uid = None

        self.verbose = verbose

        self.backgrounds = backgrounds
        self.bg_correction_enabled = True
        self.correlation = correlation
        self.corr_type = corr_type
        self.filename = filename

        self.fit_algorithm = fit_algorithm
        self.fit_ival = fit_ival
        self.fit_model = fit_model
        # Do not change order:
        self.fit_weight_type = fit_weight_type
        self.fit_weight_parameters = fit_weight_data

        self.normparm = normparm
        self.title = title
        self.traces = traces

    def __repr__(self):
        if self.is_ac:
            c = "AC"
        else:
            c = "CC"
        text = "{} correlation '{}' with {} traces".format(
            c, self.title, len(self._traces))
        return text

    def background_replace(self, channel, background):
        """
        Replace a background.
        Channel must be 0 or 1.
        background must be instance of `Trace`
        """
        if channel not in [0, 1]:
            raise ValueError("`channel` must be '0' or '1'!")
        if not isinstance(background, Trace):
            raise ValueError("`background` must be instance of `Trace`")

        if self.is_ac:
            if channel == 1:
                raise ValueError("Cannot set second background for AC.")
            self._backgrounds = [background]
        else:
            if len(self._backgrounds) == 0:
                self._backgrounds = [Trace(countrate=0, duration=0),
                                     Trace(countrate=0, duration=0)]
            elif len(self._backgrounds) == 1:
                self._backgrounds.append(Trace(countrate=0, duration=0))
            self._backgrounds[channel] = background

    @property
    def backgrounds(self):
        """
        The background trace(s) of this correlation in a list.
        """
        return self._backgrounds

    @backgrounds.setter
    def backgrounds(self, value):
        """
        Set the backgrounds. The value can either be a list of traces or
        instances of traces or a single trace in an array.
        """
        backgrounds = []
        if not isinstance(value, list):
            value = [value]
        if len(value) > 2:
            raise ValueError("Backgrounds length must not exceed 2.")

        for v in value:
            if isinstance(v, np.ndarray):
                backgrounds.append(Trace(trace=v))
            elif isinstance(v, Trace):
                backgrounds.append(v)
            else:
                raise ValueError(
                    "Each background must be instance of Trace or ndarray")
        self._backgrounds = backgrounds

    @property
    def bg_correction_factor(self):
        """
        Returns background correction factor for
        self._correlation

        Notes
        -----
        Thompson, N. Lakowicz, J.;
        Geddes, C. D. & Lakowicz, J. R. (ed.)
        Fluorescence Correlation Spectroscopy
        Topics in Fluorescence Spectroscopy,
        Springer US, 2002, 1, 337-378
        """
        if not self.bg_correction_enabled:
            # bg correction disabled
            return 1

        if self.is_ac:
            # Autocorrelation
            if len(self.traces) == 1 and len(self.backgrounds) == 1:
                S = self.traces[0].countrate
                B = self.backgrounds[0].countrate
                bgfactor = (S/(S-B))**2
            else:
                if self.verbose >= 1:
                    warnings.warn("Correlation {}: no bg-correction".
                                  format(self.uid))
                bgfactor = 1
        else:
            # Crosscorrelation
            if len(self.traces) == 2 and len(self.backgrounds) == 2:
                S = self.traces[0].countrate
                S2 = self.traces[1].countrate
                B = self.backgrounds[0].countrate
                B2 = self.backgrounds[1].countrate
                bgfactor = (S/(S-B)) * (S2/(S2-B2))
            else:
                warnings.warn("Correlation {}: no bg-correction".
                              format(self))
                bgfactor = 1
        return bgfactor

    def check_parms(self, parms):
        """ Check parameters using self.fit_model.func_verification and the user defined
            boundaries self.fit_parameters_range for each parameter.
        """
        p = 1. * np.array(parms)
        r = self.fit_parameters_range
        for i in range(len(p)):
            if r[i][0] == r[i][1]:
                pass
            elif r[i][0] is None:
                if p[i] > r[i][1]:
                    p[i] = r[i][1]
            elif r[i][1] is None:
                if p[i] < r[i][0]:
                    p[i] = r[i][1]
            elif p[i] < r[i][0]:
                p[i] = r[i][0]
            elif p[i] > r[i][1]:
                p[i] = r[i][1]
        return p

    @property
    def correlation(self):
        """the correlation data, shape (N,2) with (time, correlation) """
        if self._correlation is not None:
            corr = self._correlation.copy()
            return corr

    @correlation.setter
    def correlation(self, value):
        if value is None:
            warnings.warn("Setting correlation to `None`.")
        elif not isinstance(value, np.ndarray):
            raise ValueError("Correlation must be 2d array!")
        elif not value.shape[1] == 2:
            raise ValueError("Correlation array must have shape (N,2)!")
        self._correlation = value

    @property
    def correlation_fit(self):
        """ returns correlation data for fitting (fit_ivald)
        - background correction
        - fitting interval cropping
        """
        corr = self.correlation
        if corr is not None:
            # perform background correction
            corr[:, 1] *= self.bg_correction_factor
            # perform parameter normalization
            return corr[self.fit_ival[0]:self.fit_ival[1], :]

    @property
    def correlation_plot(self):
        """ returns correlation data for plotting (normalized, fit_ivald)
        - background correction
        - fitting interval cropping
        - parameter normalization
        """
        corr = self.correlation_fit
        if corr is not None:
            # perform parameter normalization
            corr[:, 1] *= self.normalize_factor
            return corr

    @property
    def is_ac(self):
        """True if instance contains autocorrelation"""
        return self.corr_type.lower().count("ac") > 0

    @property
    def is_cc(self):
        """True if instance contains crosscorrelation"""
        return not self.is_ac

    @property
    def is_weighted_fit(self):
        """True if a weighted fit was performed"""
        return self.fit_weight_type != "none"

    @property
    def fit_algorithm(self):
        """The string representing the fitting algorithm"""
        return self._fit_algorithm

    @fit_algorithm.setter
    def fit_algorithm(self, value):
        # TODO:
        # - allow lower-case fitting algorithm

        if value not in list(fit.Algorithms.keys()):
            raise ValueError("Invalid fit algorithm: {}".format(value))

        self._fit_algorithm = value

    @property
    def fit_ival(self):
        """lag time interval for fitting"""
        lag = self.lag_time
        if lag is not None:
            if self._fit_ival[1] <= 0 or self._fit_ival[1] > lag.shape[0]:
                self._fit_ival[1] = lag.shape[0]
        return self._fit_ival

    @fit_ival.setter
    def fit_ival(self, value):
        value = list(value)
        if value[1] <= 0:
            if self.lag_time is not None:
                value[1] = self.lag_time.shape[0]
            else:
                # just to be sure
                warnings.warn("No data available.")
                value[1] = 10000000000000000
        self._fit_ival = value

    @property
    def fit_model(self):
        """instance of a fit model"""
        return self._fit_model

    @fit_model.setter
    def fit_model(self, value):
        """set the fit model
        """
        if isinstance(value, int):
            newmodel = mdls.modeldict[value]
        elif isinstance(value, mdls.Model):
            newmodel = value
        else:
            raise NotImplementedError("Unknown model identifier")

        if newmodel != self._fit_model:
            self._fit_model = newmodel
            # overwrite fitting parameters
            self._fit_parameters = self._fit_model.default_values
            self._fit_parameters_variables = self._fit_model.default_variables
            self._fit_parameters_range = np.zeros(
                (len(self._fit_parameters), 2))
            self.normparm = None

    @property
    def fit_weight_data(self):
        """data of weighted fitting"""
        try:
            data = self._fit_weight_memory[self.fit_weight_type]
        except KeyError:
            # Standard variables for weights
            if self.fit_weight_type.count("spline"):
                # Default area for weighting with spline fit
                data = 3
            else:
                data = None
        return data

    @fit_weight_data.setter
    def fit_weight_data(self, value):
        self._fit_weight_memory[self.fit_weight_type] = value

    @property
    def fit_parameters(self):
        """parameters that were fitted/will be used for fitting"""
        # Do not return `self._fit_parameters.copy()`, because
        # some methods of PyCorrFit depend on the array being
        # accessible and changeable with indices.
        return self._fit_parameters

    @fit_parameters.setter
    def fit_parameters(self, value):
        # must unlock parameters, if change is required
        value = np.array(value)
        self._fit_parameters = self.check_parms(value)

    @property
    def fit_parameters_range(self):
        """valid fitting ranges for fit parameters"""
        model = self.fit_model.boundaries
        mine = self._fit_parameters_range
        new = []
        for a, b in zip(model, mine):
            c = [-np.inf, np.inf]
            if a[0] != a[1]:
                c[0] = a[0]
                c[1] = a[1]
            # user overrides model
            if b[0] != b[1]:
                c[0] = b[0]
                c[1] = b[1]
            if c[0] is not None and np.isnan(c[0]):
                c[0] = -np.inf
            if c[1] is not None and np.isnan(c[1]):
                c[1] = np.inf

            new.append(c)
        return np.array(new)

    @fit_parameters_range.setter
    def fit_parameters_range(self, value):
        value = np.array(value)
        expect_shape = (self.fit_parameters.shape[0], 2)
        if value.shape != expect_shape:
            msg = "Expected shape of fit parameters: {} (vs {})".format(
                expect_shape, value.shape
            )
            raise ValueError(msg)
        self._fit_parameters_range = value

    @property
    def fit_parameters_variable(self):
        """which parameters are variable during fitting"""
        if self._fit_parameters_variable is None:
            self._fit_parameters_variable = np.array(
                self.fit_model.default_variables, dtype=bool)
        return self._fit_parameters_variable

    @fit_parameters_variable.setter
    def fit_parameters_variable(self, value):
        value = np.array(value, dtype=bool)
        expect_size = self.fit_parameters.shape[0]
        if value.shape[0] != expect_size:
            msg = "Fit parameter variables must have size {}!".format(
                expect_size)
            raise ValueError(msg)
        self._fit_parameters_variable = value

    @property
    def lag_time(self):
        """logarithmic lag time axis"""
        if self.correlation is not None:
            return self._correlation[:, 0].copy()
        elif self._lag_time is not None:
            return self._lag_time
        else:
            # some default lag time
            return 10**np.linspace(-6, 8, 1001)

    @lag_time.setter
    def lag_time(self, value):
        if self.correlation is not None:
            warnings.warn(
                "Setting lag time not possible, because of existing correlation")
        else:
            self._lag_time = value

    @property
    def lag_time_fit(self):
        """lag time as used for fitting"""
        return self.lag_time[self.fit_ival[0]:self.fit_ival[1]]

    @property
    def modeled(self):
        """fitted data values, same shape as self.correlation"""
        # perform parameter normalization
        lag = self.lag_time
        modeled = np.zeros((lag.shape[0], 2))
        modeled[:, 0] = lag
        modeled[:, 1] = self.fit_model(self.fit_parameters, lag)
        return modeled.copy()

    @property
    def modeled_fit(self):
        """fitted data values, same shape as self.correlation_fit"""
        toplot = self.modeled[self.fit_ival[0]:self.fit_ival[1], :]
        return toplot

    @property
    def modeled_plot(self):
        """fitted data values, same shape as self.correlation_fit"""
        toplot = self.modeled_fit
        toplot[:, 1] *= self.normalize_factor
        return toplot

    @property
    def normalize_factor(self):
        """plot normalization according to self.normparm"""
        if self.normparm is None:
            # nothing to do
            return 1
        if self.normparm < self.fit_parameters.shape[0]:
            nfactor = self.fit_parameters[self.normparm]
        else:
            # get supplementary parameters
            alt = self.fit_model.get_supplementary_values(self.fit_parameters)
            nfactor = alt[self.normparm - self.fit_parameters.shape[0]]

        return nfactor

    @property
    def residuals(self):
        """fit residuals, same shape as self.correlation"""
        if self.correlation is None:
            raise ValueError("Cannot compute residuals; No correlation given!")
        residuals = self.correlation.copy()
        residuals[:, 1] -= self.modeled[:, 1]
        return residuals

    @property
    def residuals_fit(self):
        """fit residuals, same shape as self.correlation_fit"""
        residuals_fit = self.correlation_fit.copy()
        residuals_fit[:, 1] -= self.modeled_fit[:, 1]
        return residuals_fit

    @property
    def residuals_plot(self):
        """fit residuals, same shape as self.correlation_fit"""
        cp = self.correlation_plot
        if cp is not None:
            residuals_plot = self.correlation_plot.copy()
            residuals_plot[:, 1] -= self.modeled_plot[:, 1]
            return residuals_plot

    def set_weights(self, type_name, data):
        """
        Add weights for fitting.
        example:
        type_name : "Average"
        data : 1d ndarray with length self.lag_time
        """
        if data is not None:
            self._fit_weight_memory[type_name] = data

    @property
    def traces(self):
        """
        The trace(s) of this correlation in a list.
        """
        return self._traces

    @traces.setter
    def traces(self, value):
        """
        Set the traces. The value can either be a list of traces or
        instances of traces or a single trace in an array.
        """
        traces = []
        if not isinstance(value, list):
            value = [value]
        if len(value) > 2:
            raise ValueError("Traces length must not exceed 2.")
        for v in value:
            if isinstance(v, np.ndarray):
                traces.append(Trace(trace=v))
            elif isinstance(v, Trace):
                traces.append(v)
            else:
                raise ValueError(
                    "Each trace must be instance of Trace or ndarray")
        self._traces = traces

        if len(self._traces) == 2:
            if self._traces[0].duration != self._traces[1].duration:
                warnings.warn("Unequal lenght of traces: {} and {}".format(
                              self._traces[0].duration,
                              self._traces[1].duration))

    @property
    def uid(self):
        """
        unique identifier of this instance
        This might change when title or filename
        are updated.
        """
        if self._uid is None:
            hasher = hashlib.sha256()
            hasher.update(str(np.random.random()).encode())
            hasher.update(str(self._correlation).encode())
            hasher.update(str(self.filename).encode())
            hasher.update(str(self.title).encode())
            self._uid = hasher.hexdigest()
        return self._uid



import numpy as np

from .control import model_setup
from .cp_confocal import twod
from .cp_triplet import trip


# 2D + TT Gauß
# Model 6003
def CF_Gxy_gauss_2DTT(parms, tau):
    u""" Two-dimensional free diffusion
        with a Gaussian laser profile, including two triplet components.
        The triplet factor takes into account a blinking term.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        particle = 1/(1+τ/τ_diff)
        triplet1 = 1 + T₁/(1-T₁)*exp(-τ/τ_trip₁)
        triplet2 = 1 + T₂/(1-T₂)*exp(-τ/τ_trip₂)
        G = 1/n*particle*triplet1*triplet2 + offset

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0]  n        Effective number of particles in confocal volume
        [1]  τ_diff   Diffusion time of particle
        [2]  τ_trip₁  Characteristic residence time in triplet state
        [3]  T₁       Fraction of particles in triplet (non-fluorescent) state
                      0 <= T < 1
        [4]  τ_trip₂  Characteristic residence time in triplet state
        [5]  T₂       Fraction of particles in triplet (non-fluorescent) state
                      0 <= T < 1
        [6]  offset
        *tau* - lag time
    """
    n = parms[0]
    taud = parms[1]
    tautrip1 = parms[2]
    T1 = parms[3]
    tautrip2 = parms[4]
    T2 = parms[5]
    off = parms[6]

    g = twod(tau=tau, taudiff=taud)

    tr1 = trip(tau=tau, T=T1, tautrip=tautrip1)
    tr2 = trip(tau=tau, T=T2, tautrip=tautrip2)

    G = off + 1/n * g * tr1 * tr2

    return G


def supplements(parms, countrate=None):
    # We can only give you the effective particle number
    n = parms[0]
    Info = list()
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [
    4,       # n
    .4,      # taud
    0.001,   # tautrip1
    0.01,    # T1
    0.002,   # tautrip2
    0.01,    # T2
    0.0      # offset
]

# Boundaries
# strictly positive
boundaries = [[0, np.inf]]*len(parms)
# T
boundaries[3] = [0, .9999999999999]
boundaries[5] = [0, .9999999999999]
# offset
boundaries[-1] = [-np.inf, np.inf]


model_setup(
    modelid=6003,
    name="2D diffusion with double triplet (confocal)",
    comp="T+T+2D",
    mtype="Confocal (Gaussian) with double triplet",
    fctn=CF_Gxy_gauss_2DTT,
    par_labels=[
        u"n",
        u"τ_diff [ms]",
        u"τ_trip₁ [ms]",
        u"T₁",
        u"τ_trip₂ [ms]",
        u"T₂",
        u"offset"
    ],
    par_values=parms,
    par_vary=[True, True, False, False, False, False, False],
    par_boundaries=boundaries,
    par_constraints=[[2, "<", 1], [4, ">", 2]],
    par_hr_labels=[
        u"n",
        u"τ_diff [ms]",
        u"τ_trip₁ [µs]",
        u"T₁",
        u"τ_trip₂ [µs]",
        u"T₂",
        u"offset"
    ],
    par_hr_factors=[
        1.,     # n
        1.,     # taudiff
        1000.,  # tautrip1 [µs]
        1.,     # T1
        1000.,  # tautrip2 [µs]
        1.,     # T2
        1.      # offset
    ],
    supplementary_method=supplements
)



"""Mixed components for fitting models"""


def double_pnum(n,
                F1,
                alpha,
                comp1,
                kwargs1,
                comp2,
                kwargs2,
                ):
    u"""
    Double component models where the particle number is given in
    the model i.e. for confocal diffusion models.

    Parameters
    ----------
    n : float
        Total particle number
    F1 : float
        Fraction of particle species 1
    alpha : float
        Relative molecular brightness of particle
        2 compared to particle 1 (α = q₂/q₁)
    comp1, comp2 : callables
        The model functions for each of the components.
    kwargs1, kwargs2 : dicts
        The keyword arguments for `comp1` and `comp2`
    """
    norm = (F1 + alpha*(1-F1))**2

    g1 = F1 * comp1(**kwargs1)
    g2 = alpha**2 * (1-F1) * comp2(**kwargs2)

    G = 1/n * (g1 + g2) / norm

    return G


def triple_pnum(n,
                F1,
                F2,
                alpha21,
                alpha31,
                comp1,
                kwargs1,
                comp2,
                kwargs2,
                comp3,
                kwargs3
                ):
    u"""
    Double component models where the particle number is given in
    the model i.e. for confocal diffusion models.

    Parameters
    ----------
    n : float
        Total particle number
    F1, F2 : float
        Fraction of particle species 1 and 2.
        This infers that F3 = 1 - F1 - F2
    alpha21 : float
        Relative molecular brightness of particle
        2 compared to particle 1 (α₂₁ = q₂/q₁)
    alpha31 : float
        Relative molecular brightness of particle
        3 compared to particle 1
    comp1, comp2, comp3 : callables
        The model functions for each of the components.
    kwargs1, kwargs2, kwargs3 : dicts
        The keyword arguments for `comp1`, `comp2`, and `comp3`.
    """
    alpha11 = 1
    F3 = 1 - F1 - F2
    if F3 < 0:
        F3 = 0

    norm = (F1*alpha11 + F2*alpha21 + F3*alpha31)**2

    g1 = alpha11**2 * F1 * comp1(**kwargs1)
    g2 = alpha21**2 * F2 * comp2(**kwargs2)
    g3 = alpha31**2 * F3 * comp3(**kwargs3)

    G = 1/n * (g1 + g2 + g3) / norm

    return G



import numpy as np

from pycorrfit.correlation import Correlation
from pycorrfit.fit import Fit


def create_corr():
    corr = Correlation()

    tau = np.exp(np.linspace(np.log(1e-3), np.log(1e6), 10))
    data = corr.fit_model(corr.fit_parameters, tau)
    noise = (np.random.random(data.shape[0])-.5)*.0005
    data += noise

    corr.correlation = np.dstack((tau, data))[0]
    return corr


def test_simple_corr():
    corr = create_corr()
    oldparms = corr.fit_parameters.copy()
    temp = corr.fit_parameters
    temp[0] *= 2
    temp[-1] *= .1

    Fit(corr)

    res = oldparms - corr.fit_parameters

    assert np.allclose(res, np.zeros_like(res), atol=0.010)


if __name__ == "__main__":
    import matplotlib.pylab as plt
    corr = create_corr()

    fig, (ax1, ax2) = plt.subplots(2, 1)
    ax1.set_xscale("log")
    ax2.set_xscale("log")

    print(corr.fit_parameters)
    temp = corr.fit_parameters
    temp[0] *= 2
    temp[-1] *= .1
    ax1.plot(corr.correlation_fit[:, 0], corr.correlation_fit[:, 1])
    ax1.plot(corr.modeled_fit[:, 0], corr.modeled_fit[:, 1])
    print(corr.fit_parameters)

    Fit(corr)

    print(corr.fit_parameters)

    ax2.plot(corr.correlation_fit[:, 0], corr.correlation_fit[:, 1])
    ax2.plot(corr.modeled_fit[:, 0], corr.modeled_fit[:, 1])

    plt.show()



"""Confocal fitting model components"""
import numpy as np


def threed(tau, taudiff, SP):
    return 1/((1 + tau/taudiff) * np.sqrt(1+tau/(taudiff*SP**2)))


def twod(tau, taudiff):
    return 1/((1 + tau/taudiff))



"""Module readfiles: Import correlation data from data files"""
import csv
import io
import pathlib
import shutil
import sys
import tempfile
import warnings
import zipfile

import numpy as np
import yaml

# To add a filetype add it here and in the
# dictionaries at the end of this file.
from .read_ASC_ALV import openASC
from .read_cor_PicoQuant import openCOR
from .read_CSV_PyCorrFit import openCSV
from .read_SIN_correlator_com import openSIN
from .read_FCS_Confocor3 import openFCS
from .read_mat_ries import openMAT
from .read_pt3_PicoQuant import openPT3
from .read_ptu_PicoQuant import openPTU


def add_all_supported_filetype_entry(adict):
    wildcard = ""
    keys = adict.keys()
    N = len(keys)
    i = 0
    for key in keys:
        newwc = key.split("|")[1]
        wildcard = wildcard + newwc
        i = i + 1
        if i != N:
            wildcard = wildcard + ";"
    adict[ALL_SUP_STRING+"|"+wildcard] = open_any


def get_supported_extensions():
    """
    Returns list of extensions of currently supported file types.
    """
    extlist = []
    for kf in list(filetypes_dict.keys()):
        ext = kf.split("|")[-1]
        ext = ext.split(";")
        ext = [e.lower().strip("*. ") for e in ext]
        ext = list(np.unique(ext))
        extlist += ext
    extlist = list(np.unique(extlist))
    extlist.sort()
    return extlist


# To increase user comfort, we will now create a file opener thingy that
# knows how to open all files we know.
def open_any(path, filename=None):
    """Open a supported data file

    Parameters
    ----------
    path : str
        Full path to file or directory containing `filename`
    filename : str
        The name of the file if not given in path (optional).
    """
    path = pathlib.Path(path)
    if filename is not None:
        warnings.warn("Using `filename` is deprecated.", DeprecationWarning)
        path = path / filename
    wildcard = path.suffix
    for key in filetypes_dict.keys():
        # Recurse into the wildcards
        wildcardstring = key.split("|")
        # We do not want to recurse
        if wildcardstring[0] != ALL_SUP_STRING:
            otherwcs = wildcardstring[1].split(";")
            for wc in otherwcs:
                if wc.strip("*") == wildcard:
                    return filetypes_dict[key](path)
    else:
        # If we could not find the correct function in filetypes_dict,
        # try again in filetypes_bg_dict:
        return open_any_bg(path)


def open_any_bg(path, filename=None):
    path = pathlib.Path(path)
    if filename is not None:
        warnings.warn("Using `filename` is deprecated.", DeprecationWarning)
        path = path / filename

    wildcard = path.suffix
    for key in filetypes_bg_dict.keys():
        # Recurse into the wildcards
        wildcardstring = key.split("|")
        # We do not want to recurse
        if wildcardstring[0] != ALL_SUP_STRING:
            otherwcs = wildcardstring[1].split(";")
            for wc in otherwcs:
                if wc.strip("*") == wildcard:
                    return filetypes_bg_dict[key](path)
    else:
        # For convenience in openZIP
        return None


def openZIP(path, filename=None):
    """Load everything inside a .zip file that could be an FCS curve.

    Will use any wildcard in `filetypes_dict`.
    """
    #    It's a rather lengthy import of the session file. The code is copied
    #    from openfile.OpenSession. The usual zip file packed curves are
    #    imported on the few code lines after the else statement.

    path = pathlib.Path(path)
    if filename is not None:
        warnings.warn("Using `filename` is deprecated.", DeprecationWarning)
        path = path / filename
    filename = path.name

    # Open the archive:
    Arc = zipfile.ZipFile(path, mode='r')
    Correlations = []  # Correlation information
    Curvelist = []    # Type information
    Filelist = []     # List of filenames corresponding to *Curvelist*
    Trace = []        # Corresponding traces
    # First test, if we are opening a session file
    sessionwc = [".fcsfit-session.zip", ".pcfs"]
    if filename.endswith(sessionwc[0]) or filename.endswith(sessionwc[1]):
        # Get the yaml parms dump:
        yamlfile = Arc.open("Parameters.yaml")
        # Parms: Fitting and drawing parameters of the correlation curve
        # The *yamlfile* is responsible for the order of the Pages #i.
        # The parameters are actually useless to us right now.
        Parms = yaml.safe_load(yamlfile)
        yamlfile.close()
        # Get the correlation arrays
        ImportedNum = list()
        for i in np.arange(len(Parms)):
            # The *number* is used to identify the correct file
            number = str(Parms[i][0])
            expfilename = "data"+number[1:len(number)-2]+".csv"
            expfile = Arc.open(expfilename, 'r')
            readdata = csv.reader(io.StringIO(
                expfile.read().decode()), delimiter=',')
            dataexp = list()
            if str(readdata.__next__()[0]) == "# tau only":
                # We do not have a curve here
                pass
            else:
                Filelist.append(filename+"/#"+number[1:len(number)-2])
                for row in readdata:
                    # Exclude commentaries
                    if (str(row[0])[0:1] != '#'):
                        dataexp.append((float(row[0]), float(row[1])))
                dataexp = np.array(dataexp)
                Correlations.append(dataexp)
                ImportedNum.append(i)
            del readdata
            expfile.close()
        # Get the Traces
        for i in ImportedNum:
            # Make sure we only import those traces that had a corresponding
            # correlation curve. (ImportedNum)
            #
            # The *number* is used to identify the correct file
            number = str(Parms[i][0])
            # Find out, if we have a cross correlation data type
            IsCross = False
            try:
                IsCross = Parms[i][7]
            except IndexError:
                # No Cross correlation
                pass
            if IsCross is False:
                tracefilenames = ["trace"+number[1:len(number)-2]+".csv"]
                Curvelist.append("AC")
            else:
                # Cross correlation uses two traces
                tracefilenames = ["trace"+number[1:len(number)-2]+"A.csv",
                                  "trace"+number[1:len(number)-2]+"B.csv"]
                Curvelist.append("CC")
            thistrace = list()
            for tracefilename in tracefilenames:
                try:
                    Arc.getinfo(tracefilename)
                except KeyError:
                    # No correlation curve, but add a None
                    pass
                else:
                    tracefile = Arc.open(tracefilename, 'r')
                    traceread = csv.reader(io.StringIO(
                        tracefile.read().decode()), delimiter=',')
                    singletrace = list()
                    for row in traceread:
                        # Exclude commentaries
                        if (str(row[0])[0:1] != '#'):
                            singletrace.append((float(row[0]), float(row[1])))
                    singletrace = np.array(singletrace)
                    thistrace.append(singletrace)
                    del traceread
                    del singletrace
                    tracefile.close()
            if len(thistrace) == 1:
                Trace.append(thistrace[0])
            elif len(thistrace) == 2:
                Trace.append(thistrace)
            else:
                Trace.append(None)
    else:
        # We are not importing from a session but from a zip file with
        # probably a mix of all filetypes we know. This works
        # recursively (e.g. a zip file in a zipfile).
        allfiles = Arc.namelist()
        # Extract data to temporary folder
        tempdir = pathlib.Path(tempfile.mkdtemp())
        for afile in allfiles:
            Arc.extract(afile, path=tempdir)
            data = open_any(tempdir / afile)
            if data is not None:
                Correlations += data["Correlation"]
                Trace += data["Trace"]
                Curvelist += data["Type"]
                fnames = data["Filename"]
                Filelist += [filename+"/"+fs for fs in fnames]
        shutil.rmtree(path=tempdir, ignore_errors=True)
    Arc.close()
    dictionary = {}
    dictionary["Correlation"] = Correlations
    dictionary["Trace"] = Trace
    dictionary["Type"] = Curvelist
    dictionary["Filename"] = Filelist
    return dictionary


# The string that is shown when opening all supported files
# We add an empty space so it is listed first in the dialogs.
ALL_SUP_STRING = " All supported files"

# Dictionary with filetypes that we can open
# The wildcards point to the appropriate functions.
filetypes_dict = {"Correlator.com (*.SIN)|*.SIN;*.sin": openSIN,
                  "ALV (*.ASC)|*.ASC;*.asc": openASC,
                  "PicoQuant cor (*.cor)|*.cor": openCOR,
                  "PyCorrFit (*.csv)|*.csv": openCSV,
                  "Matlab 'Ries (*.mat)|*.mat": openMAT,
                  "PicoQuant (*.pt3)|*.pt3": openPT3,
                  "PicoQuant PTU T2/T3 (*.ptu)|*.ptu": openPTU,
                  "Zeiss ConfoCor3 (*.fcs)|*.fcs": openFCS,
                  "Zip file (*.zip)|*.zip": openZIP,
                  "PyCorrFit session (*.pcfs)|*.pcfs": openZIP
                  }
# For user comfort, add "All supported files" wildcard:
add_all_supported_filetype_entry(filetypes_dict)

# Dictionary with filetypes we can open that have intensity traces in them.
filetypes_bg_dict = {"Correlator.com (*.SIN)|*.SIN;*.sin": openSIN,
                     "ALV (*.ASC)|*.ASC": openASC,
                     "PicoQuant cor (*.cor)|*.cor": openCOR,
                     "PyCorrFit (*.csv)|*.csv": openCSV,
                     "PicoQuant (*.pt3)|*.pt3": openPT3,
                     "Zeiss ConfoCor3 (*.fcs)|*.fcs": openFCS,
                     "Zip file (*.zip)|*.zip": openZIP,
                     "PyCorrFit session (*.pcfs)|*.pcfs": openZIP
                     }
add_all_supported_filetype_entry(filetypes_bg_dict)



"""PyCorrFit - update checking"""
from looseversion import LooseVersion  # For version checking
import os
import tempfile
import traceback
import urllib.request
import webbrowser

import simplejson
import wx.html
import wx.lib.delayedresult as delayedresult

from .. import _version as pcf_version

from . import doc                          # Documentation/some texts
# The icon file was created with
# img2py -i -n Main PyCorrFit_icon.png icon.py
from . import misc


class UpdateDlg(wx.Frame):
    def __init__(self, parent, valuedict):

        description = valuedict["Description"]
        homepage = valuedict["Homepage"]
        githome = valuedict["Homepage_GIT"]
        changelog = valuedict["Changelog"]
        pos = parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent, wx.ID_ANY, title="Update",
                          size=(230, 240), pos=pos)
        self.changelog = changelog
        # Fill html content
        html = wxHTML(self)
        string = '' \
                 + "<b> PyCorrFit <br></b>" \
                 + "Your version: {}<br>".format(description[0]) \
                 + "Latest version: {}<br>".format(description[1]) \
                 + "({})<br><p><b>".format(description[2])
        if homepage:
            string = string + '<a href="{}">Homepage</a><br>'.format(homepage)
        if githome:
            string = string + '<a href="{}">Repository</a><br>'.format(githome)

        if changelog:
            string = string + \
                '<a href="{}">Change Log</a>'.format(changelog)
        string += '</b></p>'
        html.SetPage(string)
        self.Bind(wx.EVT_CLOSE, self.Close)
        # Set window icon
        ico = misc.getMainIcon()
        wx.Frame.SetIcon(self, ico)

    def Close(self, event):
        if len(self.changelog) != 0:
            # Cleanup downloaded file, if it was downloaded
            if self.changelog != doc.StaticChangeLog:
                os.remove(self.changelog)
        self.Destroy()


class wxHTML(wx.html.HtmlWindow):
    def OnLinkClicked(self, link):
        webbrowser.open(link.GetHref())


def get_gh_version(ghrepo="user/repo", timeout=20):
    """ Check GitHub repository for latest release
    """
    u = "https://api.github.com/repos/{}/releases/latest".format(ghrepo)
    try:
        data = urllib.request.urlopen(u, timeout=timeout).read()
    except:
        newversion = None
        try:
            with open("check_update_error.log", "w") as fe:
                fe.writelines(str(traceback.format_exc()))
        except:
            pass
    else:
        j = simplejson.loads(data)
        newversion = LooseVersion(j["tag_name"])

    return newversion


def update(parent):
    """ This is a thread for _Update """
    parent.StatusBar.SetStatusText("Connecting to server...")
    delayedresult.startWorker(_update_consumer, _update_worker,
                              wargs=(parent,), cargs=(parent,))


def _update_consumer(delayedresult, parent):
    results = delayedresult.get()
    dlg = UpdateDlg(parent, results)
    dlg.Show()
    parent.StatusBar.SetStatusText(
        "...update status: "+results["Description"][2])


def _update_worker(parent):
    # Online changelog file
    cl_file = doc.GitChLog
    try:
        responseVer = urllib.request.urlopen(cl_file, timeout=2)
    except:
        changelog = ""
    else:
        changelog = responseVer.read()

    ghrepo = "FCS-analysis/PyCorrFit"
    if hasattr(pcf_version, "repo_tag"):
        old = LooseVersion(pcf_version.repo_tag)
    else:
        old = LooseVersion(pcf_version.version)

    new = get_gh_version(ghrepo)

    if new is not None:
        if new > old:
            action = "update available"
        elif new < old:
            action = "ahead of release"
        else:
            action = "state of the art"
    description = [old, new, action]
    if changelog:
        changelogfile = tempfile.mktemp()+"_PyCorrFit_CHANGELOG"+".txt"
        clfile = open(changelogfile, 'wb')
        clfile.write(changelog)
        clfile.close()
    else:
        changelogfile = doc.StaticChangeLog
    results = dict()
    results["Description"] = description
    results["Homepage"] = doc.HomePage
    results["Homepage_GIT"] = doc.GitHome
    results["Changelog"] = changelogfile
    return results



"""Constraints of model functions"""
import os

import numpy as np
import pytest

import data_file_dl
import pycorrfit as pcf

NOAPITOKEN = "GITHUB_API_TOKEN" not in os.environ

examplefile = "Zeiss_Confocor3_LSM780_FCCS_HeLa_2015/019_cp_KIND+BFA.fcs"


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_fit_constraint_simple_inequality():
    """ Check "smaller than" relation during fitting.
    """
    dfile = data_file_dl.get_data_file(examplefile)
    data = pcf.readfiles.open_any(dfile)
    corr = pcf.Correlation(correlation=data["Correlation"][0],
                           traces=data["Trace"][0],
                           corr_type=data["Type"][0],
                           filename=os.path.basename(dfile),
                           title="test correlation",
                           fit_model=6035  # confocal 3D+3D)
                           )
    corr.fit_parameters_variable = [True, True, True, True,
                                    False, False, False]
    # crop triplet data
    corr.fit_ival[0] = 8
    pcf.Fit(corr)
    assert corr.fit_parameters[1] <= corr.fit_parameters[2]
    # -> deliberately reverse everything and try again
    corr.fit_parameters[1], corr.fit_parameters[2] = (corr.fit_parameters[2],
                                                      corr.fit_parameters[1])
    corr.fit_parameters[3] = 1-corr.fit_parameters[3]
    pcf.Fit(corr)
    # This tests also for equality
    assert corr.fit_parameters[1] <= corr.fit_parameters[2]
    if corr.fit_parameters[1] == corr.fit_parameters[2]:
        print("found identity of fit parameters - \
multiplying by two to see if relation holds")
        corr.fit_parameters[2] *= 2
        pcf.Fit(corr)
        assert corr.fit_parameters[1] < corr.fit_parameters[2]


@pytest.mark.xfail(NOAPITOKEN, reason="Restrictions to GitHub API")
@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_fit_constraint_sum_smaller_one():
    """ Check "a+b<c" relation during fitting.
    """
    dfile = data_file_dl.get_data_file(examplefile)
    data = pcf.readfiles.open_any(dfile)
    corr = pcf.Correlation(correlation=data["Correlation"][0],
                           traces=data["Trace"][0],
                           corr_type=data["Type"][0],
                           filename=os.path.basename(dfile),
                           title="test correlation",
                           fit_model=6081  # confocal 3D+3D)
                           )
    pcf.Fit(corr)
    assert corr.fit_parameters[4] + corr.fit_parameters[5] < 1
    parms0 = np.array([
        1.13827592342,  # n
        3.0918704e-05,  # τ₁ [ms]
        1.98835792339,  # τ₂ [ms]
        2000.0,  # τ₃ [ms]
        0.972264423555,  # F₁
        0.021400173882,  # F₂
        5.0,  # SP
        1.0,  # α₂₁
        1.0,  # α₃₁
        1e-08,  # τ_trip [ms]
        0.0,  # T
        0.0,  # offset
    ])
    corr.fit_parameters = parms0

    vary = [False] * 12
    vary[4] = vary[5] = True
    corr.fit_parameters_variable = vary
    # crop triplet data
    pcf.Fit(corr)
    assert corr.fit_parameters[4] + corr.fit_parameters[5] < 1
    # -> deliberately reverse everything and try again
    corr.fit_parameters[4], corr.fit_parameters[5] = (corr.fit_parameters[5],
                                                      corr.fit_parameters[4])
    pcf.Fit(corr)
    # This tests also for equality
    assert corr.fit_parameters[4] + corr.fit_parameters[5] < 1


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import wx
from wx.lib import sized_controls
import wx.lib.agw.hyperlink as hl


class ContributeDialog(sized_controls.SizedDialog):

    def __init__(self, *args, **kwargs):
        super(ContributeDialog, self).__init__(title="Contribute to PyCorrFit",
                                               *args, **kwargs)
        pane = self.GetContentsPane()

        pane_btns = sized_controls.SizedPanel(pane)
        pane_btns.SetSizerType('vertical')
        pane_btns.SetSizerProps(align="center")

        wx.StaticText(pane_btns, label=contribute_text)
        wx.StaticText(pane_btns, label="\nLinks:")
        for ii, link in enumerate(contribute_links):
            hl.HyperLinkCtrl(pane_btns,
                             -1,
                             "[{}]  {}".format(ii+1, link),
                             URL=link)
        wx.StaticText(pane_btns, label="\n")

        button_ok = wx.Button(pane_btns, label='OK')
        button_ok.Bind(wx.EVT_BUTTON, self.on_button)
        button_ok.SetFocus()
        button_ok.SetSizerProps(expand=True)

        self.Fit()

    def on_button(self, event):
        if self.IsModal():
            self.EndModal(event.EventObject.Id)
        else:
            self.Close()


contribute_text = """
PyCorrFit has no funding and a vanishingly small developer community.
My personal objective is to keep PyCorrFit operational on Linux and
Windows which is currently limited by the free time I have available.

An active community is very important for an open source project such
as PyCorrFit. You can help this community grow (and thus help improve
PyCorrFit) in numerous ways:

1. \tTell your colleagues and peers about PyCorrFit. One of them
   \tmight be able to contribute to the project.

2. \tIf you need a new feature in PyCorrFit, publicly announce a
   \tbounty for its implementation.

3. \tIf your research heavily relies on FCS, please consider
   \tdiverting some of your resources to the development
   \tof PyCorrFit.

4. \tYou don't have to be a Python programmer to contribute. If
   \tyou are familiar with reStrucuredText or LaTeX, you might
   \tbe able to help out with the online documentation.

5. \tPlease cite: Müller et al. Bioinformatics 30(17): 2532–2533,
   \t2014 [1].

6. \tSponsor me on GitHub [2] or donate via Liberapay [3].

If you are planning to contribute to PyCorrFit, please contact me via
the PyCorrFit issue page on GitHub such that we may coordinate a pull
request.

Thank you!
Paul Müller (October 2019)
"""


contribute_links = [
    "https://dx.doi.org/10.1093/bioinformatics/btu328",
    "https://github.com/sponsors/paulmueller",
    "https://liberapay.com/paulmueller",
]



"""Module tools - example
This is an example tool. You will need to edit __init__.py inside this
folder to activate it.
Add the filename (*example*) and class (*Tool*) to either of the lists
*ImpA*  or *ImpB* in __init__.py.
"""
import wx


class Tool(wx.Frame):
    # This tool is derived from a wx.frame.
    def __init__(self, parent):
        # parent is the main frame of PyCorrFit
        self.parent = parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=self.parent, title="Example tool",
                          pos=pos, style=wx.DEFAULT_FRAME_STYLE | wx.FRAME_FLOAT_ON_PARENT)
        # MYID
        # This ID is given by the parent for an instance of this class
        self.MyID = None
        # Page - the currently active page of the notebook.
        self.Page = self.parent.notebook.GetCurrentPage()
        # Content
        self.panel = wx.Panel(self)
        btncopy = wx.Button(self.panel, wx.ID_ANY, 'Example button')
        # Binds the button to the function - close the tool
        self.Bind(wx.EVT_BUTTON, self.OnClose, btncopy)
        self.topSizer = wx.BoxSizer(wx.VERTICAL)
        self.topSizer.Add(btncopy)
        self.panel.SetSizer(self.topSizer)
        self.topSizer.Fit(self)
        self.SetMinSize(self.topSizer.GetMinSize())
        # Icon
        if parent.MainIcon is not None:
            wx.Frame.SetIcon(self, parent.MainIcon)
        self.Show(True)

    def OnClose(self, event=None):
        # This is a necessary function for PyCorrFit.
        # Do not change it.
        self.parent.toolmenu.Check(self.MyID, False)
        self.parent.ToolsOpen.__delitem__(self.MyID)
        self.Destroy()

    def OnPageChanged(self, page, trigger=None):
        """
            This function is called, when something in the panel
            changes. The variable `trigger` is used to prevent this
            function from being executed to save stall time of the user.
            Forr a list of possible triggers, see the doc string of
            `tools`.
        """
        # When parent changes
        # This is a necessary function for PyCorrFit.
        # This is stuff that should be done when the active page
        # of the notebook changes.
        if self.parent.notebook.GetPageCount() == 0:
            # Do something when there are no pages left.
            self.panel.Disable()
            return
        self.panel.Enable()
        self.Page = page



"""PyCorrFit - module "models"

Define all models and set initial parameters.

Each model has a unique ID. This ID is very important:
    1. It is a wxWidgets ID.
    2. It is used in the saving of sessions to identify a model.
It is very important, that model IDs do NOT change in newer versions
of PyCorrFit, because it would not be possible to restore older
PyCorrFit sessions.

Dimensionless representation:
unit of time        : 1 ms
unit of inverse time: 10³ /s
unit of distance    : 100 nm
unit of Diff.coeff  : 10 µm²/s
unit of inverse area: 100 /µm²
unit of inv. volume : 1000 /µm³
"""
import copy
import sys
import warnings

import numpy as np

from .classes import Model
from .control import values, valuedict, models, modeldict, modeltypes, supplement, boundaries, shorttype


def GetHumanReadableParms(model, parameters):
    """ From a set of parameters that have internal units e.g. [100 nm],
        Calculate the parameters in human readable units e.g. [nm].
        Uses modeldict from this module.
        *model* - an integer ID of a model
        *parameters* - a list/array of parameters
                       (all parameters of that model)
        Returns:
        New Units, New Parameters
    """
    stdparms = valuedict[model]
    if len(stdparms) == 5:
        # This means we have extra information on the model
        # Return some human readable stuff
        OldParameters = 1.*np.array(parameters)
        Facors = 1.*np.array(stdparms[4])
        NewParameters = 1.*OldParameters*Facors
        NewUnits = stdparms[3]
        return NewUnits, NewParameters
    else:
        # There is no info about human readable stuff, or it is already human
        # readable.
        return stdparms[0], parameters


def GetHumanReadableParameterDict(model, names, parameters):
    """ From a set of parameters that have internal units e.g. [100 nm],
        Calculate the parameters in human readable units e.g. [nm].
        Uses modeldict from this module.
        In contrast to *GetHumanReadableParms* this function accepts
        single parameter names and does not need the full array of
        parameters.
        *model* - an integer ID of a model
        *name* - the names of the parameters to be translated,
                 order should be same as in
        *parameters* - a list of parameters
        Returns:
        New Units, New Parameters
    """
    stdparms = valuedict[model]
    if len(stdparms) == 5:
        # This means we have extra information on the model
        # Return some human readable stuff
        # Check for list:
        if isinstance(names, str):
            names = [names]
            parameters = [parameters]
            retstring = True
        else:
            retstring = False
        # Create new lists
        NewUnits = list()
        NewParameters = list()
        for i in np.arange(len(stdparms[0])):
            for j in np.arange(len(names)):
                if names[j] == stdparms[0][i]:
                    NewUnits.append(stdparms[3][i])
                    NewParameters.append(stdparms[4][i]*parameters[j])
        if retstring == True:
            NewUnits = NewUnits[0]
            NewParameters = NewParameters[0]
        return NewUnits, NewParameters
    else:
        # There is no info about human readable stuff, or it is already human
        # readable.
        return names, parameters


def GetInternalFromHumanReadableParm(model, parameters):
    """ This is the inverse of *GetHumanReadableParms*
        *model* - an integer ID of a model
        *parameters* - a list/array of parameters
        Returns:
        New Units, New Parameters
    """
    stdparms = valuedict[model]
    if len(stdparms) == 5:
        # This means we have extra information on the model
        # and can convert to internal values
        OldParameters = 1.*np.array(parameters)
        Facors = 1./np.array(stdparms[4])        # inverse
        NewParameters = 1.*OldParameters*Facors
        NewUnits = stdparms[0]
        return NewUnits, NewParameters
    else:
        # There is no info about human readable stuff. The given
        # parameters have not been converted befor using
        # *GetHumanReadableParms*.
        return stdparms[0], parameters


def GetModelType(modelid):
    """ Given a modelid, get the type of model function
        (Confocal, TIR-Conf., TIR-□, User)
    """
    if modelid >= 7000:
        return u"User"
    else:
        for key in modeltypes.keys():
            mlist = modeltypes[key]
            if mlist.count(modelid) == 1:
                try:
                    return shorttype[key]
                except:
                    warnings.warn("No shorttype defined for `{}`.".format(key))
                    return key


def GetModelFunctionFromId(modelid):
    return modeldict[modelid][3]


def GetModelParametersFromId(modelid):
    return valuedict[modelid][1]


def GetModelFitBoolFromId(modelid):
    return valuedict[modelid][2]


def GetMoreInfo(modelid, Page):
    """ This functino is called by someone who has already calculated
        some stuff or wants to know more about the model he is looking at.
        *modelid* is an ID of a model.
        *Page* is a wx.flatnotebook page.
        Returns:
         More information about a model in form of a list.
    """
    # Background signal average
    bgaverage = None
    # Get the parameters from the current page.
    parms = Page.active_parms[1]
    Info = list()
    corr = Page.corr
    if corr.is_ac:
        if len(corr.traces) == 1:
            countrate = corr.traces[0].countrate
        else:
            countrate = None
        # First import the supplementary parameters of the model
        # The order is important for plot normalization and session
        # saving as of version 0.7.8
        # Try to get the dictionary entry of a model
        # Background information
        if len(corr.backgrounds) == 1:
            bgaverage = corr.backgrounds[0].countrate
            # Now set the correct countrate
            # We already printed the countrate, so there's no harm done.
        if countrate is not None and bgaverage is not None:
            # might be that there is no countrate.
            relativecountrate = countrate - bgaverage
        else:
            relativecountrate = countrate
        # In case of cross correlation, we don't show this kind of
        # information.
        try:
            # This function should return all important information
            # that can be calculated from the given parameters.
            # We need the relativecountrate to compute the CPP.
            func_info = supplement[modelid]
            data = func_info(parms, relativecountrate)
            for item in data:
                Info.append([item[0], item[1]])
        except KeyError:
            # No information available
            pass
        if countrate is not None:
            # Measurement time
            duration = corr.traces[0].duration/1000
            Info.append(["duration [s]", duration])
            # countrate has to be printed before background.
            # Background might overwrite countrate.
            Info.append(["avg. signal [kHz]", corr.traces[0].countrate])
    else:
        # Cross correlation curves usually have two traces. Since we
        # do not know how to compute the cpp, we will pass the argument
        # "None" as the countrate.
        # First import the supplementary parameters of the model
        # The order is important for plot normalization and session
        # saving as of version 0.7.8
        # Try to get the dictionary entry of a model
        try:
            # This function should return all important information
            # that can be calculated from the given parameters.
            func_info = supplement[modelid]
            data = func_info(parms, None)
            for item in data:
                Info.append([item[0], item[1]])
        except KeyError:
            # No information available
            pass
        if len(corr.traces) == 2:
            # Measurement time
            duration = corr.traces[0].duration/1000
            Info.append(["duration [s]", duration])
            # countrate has to be printed before background.
            # Background might overwrite countrate.
            Info.append(["avg. signal A [kHz]", corr.traces[0].countrate])
            Info.append(["avg. signal B [kHz]", corr.traces[1].countrate])

    if len(Info) == 0:
        # If nothing matched until now:
        return None
    else:
        return Info


def GetPositionOfParameter(model, name):
    """ Returns an integer corresponding to the position of the label
        of a parameter in the model function
    """
    stdparms = valuedict[model]
    for i in np.arange(len(stdparms[0])):
        if name == stdparms[0][i]:
            return int(i)



"""Module tools - background: perform background correction"""
import os
import sys
import traceback

import numpy as np
import wx
from wx.lib.agw import floatspin
import wx.lib.plot as plot

from pycorrfit import readfiles
from pycorrfit import Trace

from .. import misc

# Menu entry name
MENUINFO = ["&Background correction", "Open a file for background correction."]


class BackgroundCorrection(wx.Frame):
    def __init__(self, parent):
        self.MyName = "BACKGROUND"
        # Parent is main frame
        self.parent = parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=parent, title="Background correction",
                          pos=pos, style=wx.DEFAULT_FRAME_STYLE | wx.FRAME_FLOAT_ON_PARENT)
        # MYID
        # This ID is given by the parent for an instance of this class
        self.MyID = None
        # Current trace we are looking at
        self.activetrace = None
        # Importet trace
        self.trace = None
        # Importet trace after user decides to cange radio buttons
        self.oldtrace = None
        self.oldfilename = None
        self.average = None
        # Start drawing
        # Splitter Window
        self.sp = wx.SplitterWindow(self, style=wx.SP_NOBORDER)
        # Controls
        panel = wx.Panel(self.sp)
        # text1
        backgroundinit = (
            "Correct the amplitude for non-correlated background.\n" +
            "The background intensity <B> can be either imported\n" +
            "from a blank measurement or set manually.")
        textinit = wx.StaticText(panel, label=backgroundinit)
        # Radio buttons
        self.rbtnfile = wx.RadioButton(panel, -1, 'Blank measurement: ',
                                       style=wx.RB_GROUP)
        self.rbtnfile.SetValue(True)
        self.btnbrowse = wx.Button(panel, wx.ID_ANY, 'Browse ...')
        self.rbtnhand = wx.RadioButton(panel, -1, 'Manual, <B> [kHz]: ')
        # Spincontrol
        self.spinctrl = floatspin.FloatSpin(panel, digits=4, min_val=0,
                                            increment=.01)
        self.spinctrl.Enable(False)
        # Verbose text
        self.textfile = wx.StaticText(panel,
                                      label="No blank measurement file selected.")
        textmeanavg = wx.StaticText(panel,
                                    label="Average background signal [kHz]: ")
        self.textmean = wx.StaticText(panel, label="")
        # name
        textname = wx.StaticText(panel, label="User defined background name: ")
        sizeTextn = textname.GetSize()[0]
        self.bgname = wx.TextCtrl(panel, value="", size=(sizeTextn, -1))
        self.bgname.Enable(False)
        self.btnimport = wx.Button(panel, wx.ID_ANY, 'Import into session')
        self.btnimport.Enable(False)
        # Dropdown
        self.BGlist = ["File/User"]  # updated by self.UpdateDropdown()
        textdropdown = wx.StaticText(panel, label="Show background: ")
        self.dropdown = wx.ComboBox(panel, -1, "File/User", (15, -1),
                                    wx.DefaultSize, self.BGlist, wx.CB_DROPDOWN | wx.CB_READONLY)
        self.UpdateDropdown()
        # Radio buttons Channel1 and 2
        self.rbtnCh1 = wx.RadioButton(panel, -1, 'Ch1 ',
                                      style=wx.RB_GROUP)
        self.rbtnCh1.SetValue(True)
        self.rbtnCh2 = wx.RadioButton(panel, -1, 'Ch2')
        # Apply buttons
        self.btnapply = wx.Button(panel, wx.ID_ANY, 'Apply')
        textor = wx.StaticText(panel, label=" or ")
        self.btnrem = wx.Button(panel, wx.ID_ANY, 'Dismiss')
        textpages = wx.StaticText(panel, label=" correction for pages: ")
        self.WXTextPages = wx.TextCtrl(panel, value="")
        # Initial value for WXTextPages
        pagenumlist = list()
        for i in np.arange(self.parent.notebook.GetPageCount()):
            Page = self.parent.notebook.GetPage(i)
            pagenumlist.append(
                int("".join(filter(lambda x: x.isdigit(), Page.counter))))
        valstring = misc.parsePagenum2String(pagenumlist)
        self.WXTextPages.SetValue(valstring)
        textyma = wx.StaticText(panel, label="Shortcut - ")
        self.btnapplyall = wx.Button(panel, wx.ID_ANY, 'Apply to all pages')
        textor2 = wx.StaticText(panel, label=" or ")
        self.btnremyall = wx.Button(panel, wx.ID_ANY, 'Dismiss from all pages')
        # Bindings
        self.Bind(wx.EVT_BUTTON, self.OnBrowse, self.btnbrowse)
        self.Bind(wx.EVT_RADIOBUTTON, self.OnRadioFile, self.rbtnfile)
        self.Bind(wx.EVT_RADIOBUTTON, self.OnRadioHand, self.rbtnhand)
        self.Bind(wx.EVT_SPINCTRL, self.SpinCtrlChange, self.spinctrl)
        self.Bind(wx.EVT_BUTTON, self.OnImport, self.btnimport)
        self.Bind(wx.EVT_COMBOBOX, self.OnDraw, self.dropdown)
        self.Bind(wx.EVT_BUTTON, self.OnApply, self.btnapply)
        self.Bind(wx.EVT_BUTTON, self.OnApplyAll, self.btnapplyall)
        self.Bind(wx.EVT_BUTTON, self.OnRemove, self.btnrem)
        self.Bind(wx.EVT_BUTTON, self.OnRemoveAll, self.btnremyall)
        # Sizers
        topSizer = wx.BoxSizer(wx.VERTICAL)
        text1sizer = wx.BoxSizer(wx.HORIZONTAL)
        text1sizer.Add(self.rbtnfile)
        text1sizer.Add(self.btnbrowse)
        text2sizer = wx.BoxSizer(wx.HORIZONTAL)
        text2sizer.Add(self.rbtnhand)
        text2sizer.Add(self.spinctrl)
        textmeansizer = wx.BoxSizer(wx.HORIZONTAL)
        textmeansizer.Add(textmeanavg)
        textmeansizer.Add(self.textmean)
        dropsizer = wx.BoxSizer(wx.HORIZONTAL)
        dropsizer.Add(textdropdown)
        droprightsizer = wx.BoxSizer(wx.VERTICAL)
        dropsizer.Add(droprightsizer)
        droprightsizer.Add(self.dropdown)
        # droprightsizer.Add(self.textafterdropdown)
        applysizer = wx.BoxSizer(wx.HORIZONTAL)
        applysizer.Add(self.btnapply)
        applysizer.Add(textor)
        applysizer.Add(self.btnrem)
        applysizer.Add(textpages)
        applysizer.Add(self.WXTextPages)
        applysizer.Add(self.rbtnCh1)
        applysizer.Add(self.rbtnCh2)
        allsizer = wx.BoxSizer(wx.HORIZONTAL)
        allsizer.Add(textyma)
        allsizer.Add(self.btnapplyall)
        allsizer.Add(textor2)
        allsizer.Add(self.btnremyall)

        topSizer.Add(textinit)
        topSizer.Add(text1sizer)
        topSizer.Add(text2sizer)
        topSizer.Add(self.textfile)
        topSizer.Add(textmeansizer)
        topSizer.Add(textname)
        topSizer.Add(self.bgname)
        topSizer.Add(self.btnimport)
        topSizer.Add(dropsizer)
        topSizer.Add(applysizer)
        topSizer.Add(allsizer)
        panel.SetSizer(topSizer)
        topSizer.Fit(self)
        self.SetMinSize(topSizer.GetMinSize())
        self.Show(True)
        # Canvas
        self.canvas = plot.PlotCanvas(self.sp)
        # Sizes
        psize = panel.GetBestSize()
        initial_size = (psize[0], psize[1]+200)
        self.SetSize(initial_size)
        sashsize = psize[1]+3
        # This is also necessary to prevent unsplitting
        self.sp.SetMinimumPaneSize(sashsize)
        self.sp.SplitHorizontally(panel, self.canvas, sashsize)
        # If there is no page, disable ourselves:
        self.OnPageChanged(self.parent.notebook.GetCurrentPage())
        # Icon
        if parent.MainIcon is not None:
            wx.Frame.SetIcon(self, parent.MainIcon)

    def Apply(self, Page, backgroundid):
        if self.rbtnCh1.GetValue() == True:
            Page.bgselected = backgroundid
        else:
            Page.bg2selected = backgroundid
        if Page.IsCrossCorrelation is False:
            # Autocorrelation only has one background!
            Page.bg2selected = None

    def OnApply(self, event):
        strFull = self.WXTextPages.GetValue()
        PageNumbers = misc.parseString2Pagenum(self, strFull)
        if PageNumbers is None:
            # Something went wrong and parseString2Pagenum already displayed
            # an error message.
            return
        # BG number
        item = self.dropdown.GetSelection()
        # Apply to corresponding pages
        for i in np.arange(self.parent.notebook.GetPageCount()):
            Page = self.parent.notebook.GetPage(i)
            j = "".join(filter(lambda x: x.isdigit(), Page.counter))
            if int(j) in PageNumbers:
                self.Apply(Page, item)
                Page.OnAmplitudeCheck("init")
                Page.PlotAll()
        # Clean up unused backgrounds
        CleanupAutomaticBackground(self.parent)

    def OnApplyAll(self, event):
        self.btnrem.Enable(True)
        self.btnremyall.Enable(True)
        N = self.parent.notebook.GetPageCount()
        item = self.dropdown.GetSelection()
        for i in np.arange(N):
            # Set Page
            Page = self.parent.notebook.GetPage(i)
            try:
                self.Apply(Page, item)
                Page.OnAmplitudeCheck("init")
                Page.PlotAll()
            except OverflowError:
                errstr = "Could not apply background to Page "+Page.counter +\
                    ". \n Check the value of the trace average and the background."
                dlg = wx.MessageDialog(self, errstr, "Error",
                                       style=wx.ICON_ERROR | wx.OK | wx.STAY_ON_TOP)
                dlg.ShowModal()
                Page.bgselected = None
                Page.bg2selected = None
        # Clean up unused backgrounds
        CleanupAutomaticBackground(self.parent)

    def OnClose(self, event=None):
        self.parent.toolmenu.Check(self.MyID, False)
        self.parent.ToolsOpen.__delitem__(self.MyID)
        self.Destroy()

    def OnBrowse(self, event):
        # filetypes_bg_dict is a dictionary with filetypes that have some
        # trace signal information.
        SupFiletypes = list(readfiles.filetypes_bg_dict.keys())
        SupFiletypes.sort()
        filters = ""
        for i in np.arange(len(SupFiletypes)):
            # Add to the filetype filter
            filters = filters+SupFiletypes[i]
            if i+1 != len(SupFiletypes):
                # Add a separator
                filters = filters+"|"
        dlg = wx.FileDialog(self, "Choose a data file",
                            self.parent.dirname, "", filters, wx.FD_OPEN)
        if dlg.ShowModal() == wx.ID_OK:
            # Workaround since 0.7.5
            (dirname, filename) = os.path.split(dlg.GetPath())
            # filename = dlg.GetFilename()
            # dirname = dlg.GetDirectory()
            # Set parent dirname for user comfort
            self.parent.dirname = dirname
            try:
                # [data, trace, curvelist]
                stuff = readfiles.open_any_bg(dirname, filename)
            except:
                # The file does not seem to be what it seems to be.
                info = sys.exc_info()
                errstr = "Unknown file format:\n"
                errstr += str(filename)+"\n\n"
                errstr += str(info[0])+"\n"
                errstr += str(info[1])+"\n"
                for tb_item in traceback.format_tb(info[2]):
                    errstr += tb_item
                wx.MessageDialog(self, errstr, "Error",
                                 style=wx.ICON_ERROR | wx.OK | wx.STAY_ON_TOP)
                return
            # Usually we will get a bunch of traces. Let the user select which
            # one to take.
            if len(stuff["Filename"]) > 1:
                choices = list()
                for i2 in np.arange(len(stuff["Filename"])):
                    choices.append(str(i2)+". " + stuff["Filename"][i2] + " " +
                                   stuff["Type"][i2])
                dlg = wx.SingleChoiceDialog(self, "Choose a curve",
                                            "Curve selection", choices=choices)
                if dlg.ShowModal() == wx.ID_OK:
                    selindex = dlg.GetSelection()
                else:
                    return
            else:
                selindex = 0
            # If we accidentally recorded a cross correlation curve
            # as the background, let the user choose which trace he wants:
            channelindex = None
            if (len(stuff["Type"][selindex]) >= 2 and
                    stuff["Type"][selindex][0:2] == "CC"):
                choices = ["Channel 1", "Channel 2"]
                label = "From which channel do you want to use the trace?"
                dlg = wx.SingleChoiceDialog(self, label,
                                            "Curve selection", choices=choices)
                if dlg.ShowModal() == wx.ID_OK:
                    channelindex = dlg.GetSelection()
                    trace = stuff["Trace"][selindex][channelindex]
                else:
                    return
            else:
                trace = stuff["Trace"][selindex]
            if trace is None:
                print("WARNING: I did not find any trace data.")
                return
            # Display filename and some of the directory
            self.textfile.SetLabel("File: ..."+dirname[-10:]+"/"+filename)
            name = str(selindex)+". "+stuff["Filename"][selindex]+" " +\
                stuff["Type"][selindex]
            if channelindex is not None:
                name += " "+str(channelindex+1)
            self.bgname.SetValue(name)

            self.trace = trace
            # Calculate average
            self.average = self.trace[:, 1].mean()
            # Display average
            self.textmean.SetLabel(str(self.average)+" kHz")
            self.spinctrl.SetValue(self.average)
            # Let the user see the opened file
            self.dropdown.SetSelection(0)
            # show trace
            self.OnDraw()
            # Enable button and editable name
            self.bgname.Enable(True)
            self.btnimport.Enable(True)
        else:
            # User pressed "Abort" - do nothing.
            self.parent.dirname = dlg.GetDirectory()
            dlg.Destroy()
            return

    def OnDraw(self, event=None):
        item = self.dropdown.GetSelection()
        if item < 0:
            # Disable Apply Buttons
            self.btnapply.Enable(False)
            self.btnapplyall.Enable(False)
            # Draw the trace that was just imported
            if self.trace and self.trace.all():  # It is enougth to check that you have an array
                # Calculate average
                self.average = self.trace[:, 1].mean()
                self.activetrace = self.trace
                # self.textafterdropdown.SetLabel(" Avg:  "+str(self.average)+
                #                                " kHz")
                self.textmean.SetLabel(str(self.average))
                self.spinctrl.SetValue(self.average)
            else:
                # Clear the canvas. Looks better.
                self.canvas.Clear()
                # Don't show the average
                # self.textafterdropdown.SetLabel("")
                self.textmean.SetLabel("")
                return
        else:
            # Enable Apply Buttons
            self.btnapply.Enable(True)
            self.btnapplyall.Enable(True)
            # Draw a trace from the list
            self.activetrace = self.parent.Background[item-1].trace
        # We want to have the trace in [s] here.
        trace = 1.*self.activetrace
        trace[:, 0] = trace[:, 0]/1000
        linesig = plot.PolyLine(trace, legend='', colour='blue', width=1)
        self.canvas.Draw(plot.PlotGraphics([linesig],
                                           xLabel='time [s]',
                                           yLabel='background signal [kHz]'))

    def OnImport(self, event):
        self.parent.Background.append(
            Trace(trace=self.trace, name=self.bgname.GetValue()))
        # Next two lines are taken care of by UpdateDropdown
        # name = "{} ({:.2f} kHz)".format(self.bgname.GetValue(), self.average)
        # self.BGlist.append(name)
        self.UpdateDropdown()
        self.btnremyall.Enable(True)
        self.btnrem.Enable(True)
        self.btnapplyall.Enable(True)
        self.btnapply.Enable(True)
        self.OnDraw()
        # Update BG dropdown of each page
        for i in np.arange(self.parent.notebook.GetPageCount()):
            self.parent.notebook.GetPage(i).OnAmplitudeCheck()

    def OnPageChanged(self, page=None, trigger=None):
        """
            This function is called, when something in the panel
            changes. The variable `trigger` is used to prevent this
            function from being executed to save stall time of the user.
            Forr a list of possible triggers, see the doc string of
            `tools`.
        """
        # We do not need the *Range* Commands here yet.
        # We open and close the SelectChannelsFrame every time we
        # import some data.
        if trigger in ["parm_batch", "fit_batch", "page_add_batch"]:
            return
        if len(self.parent.Background) == 0:
            self.BGlist = list()
            self.UpdateDropdown()
            self.dropdown.SetValue("File/User")
        if self.parent.notebook.GetPageCount() == 0:
            self.sp.Disable()
            return
        self.sp.Enable()
        if len(self.BGlist) <= 0:
            self.btnrem.Enable(False)
            self.btnremyall.Enable(False)
            self.btnapply.Enable(False)
            self.btnapplyall.Enable(False)
        else:
            self.btnrem.Enable(True)
            self.btnremyall.Enable(True)
            self.btnapply.Enable(True)
            self.btnapplyall.Enable(True)
        if (self.WXTextPages.GetValue() == ""
                and self.parent.notebook.GetPageCount() != 0):
            # Initial value for WXTextPages
            pagenumlist = list()
            for i in np.arange(self.parent.notebook.GetPageCount()):
                Page = self.parent.notebook.GetPage(i)
                pagenumlist.append(
                    int("".join(filter(lambda x: x.isdigit(), Page.counter))))
            valstring = misc.parsePagenum2String(pagenumlist)
            self.WXTextPages.SetValue(valstring)

    def OnRadioFile(self, event):
        # Do not let the user change the spinctrl
        # setting.
        self.spinctrl.Enable(False)
        self.btnbrowse.Enable(True)
        # Restor the old trace
        self.trace = self.oldtrace
        if self.oldfilename is not None:
            self.textfile.SetLabel(self.oldfilename)
        if self.trace is None:
            # Disable button and editable name
            self.bgname.Enable(False)
            self.btnimport.Enable(False)
        # Let us draw
        self.dropdown.SetSelection(0)
        self.OnDraw()

    def OnRadioHand(self, event):
        # Let user enter a signal.
        self.spinctrl.Enable(True)
        self.btnbrowse.Enable(False)
        # save the old trace. We might want to switch back to it.
        if self.trace is not None:
            self.oldtrace = 1.*self.trace
            self.oldfilename = self.textfile.GetLabel()
        self.SpinCtrlChange()
        # Do not show the filename
        self.textfile.SetLabel("No file selected.")
        # Enable button and editable name
        self.bgname.Enable(True)
        self.btnimport.Enable(True)
        if len(self.bgname.GetValue()) == 0:
            # Enter something as name
            self.bgname.SetValue("User")

    def OnRemove(self, event):
        strFull = self.WXTextPages.GetValue()
        PageNumbers = misc.parseString2Pagenum(self, strFull)
        if PageNumbers is None:
            # Something went wrong and parseString2Pagenum already displayed
            # an error message.
            return
        # BG number
        # item = self.dropdown.GetSelection()
        # Apply to corresponding pages
        for i in np.arange(self.parent.notebook.GetPageCount()):
            Page = self.parent.notebook.GetPage(i)
            j = "".join(filter(lambda x: x.isdigit(), Page.counter))
            if int(j) in PageNumbers:
                if self.rbtnCh1.GetValue() == True:
                    Page.bgselected = None
                else:
                    Page.bg2selected = None
                Page.bgselected = None
                Page.OnAmplitudeCheck("init")
                Page.PlotAll()
        # Clean up unused backgrounds
        CleanupAutomaticBackground(self.parent)

    def OnRemoveAll(self, event):
        N = self.parent.notebook.GetPageCount()
        for i in np.arange(N):
            Page = self.parent.notebook.GetPage(i)
            Page.bgselected = None
            Page.bg2selected = None
            Page.OnAmplitudeCheck("init")
            Page.PlotAll()
        # Clean up unused backgrounds
        CleanupAutomaticBackground(self.parent)

    def SetPageNumbers(self, pagestring):
        self.WXTextPages.SetValue(pagestring)

    def SpinCtrlChange(self, event=None):
        # Let user see the continuous trace we will generate
        self.average = self.spinctrl.GetValue()
        self.trace = np.array([[0, self.average], [1, self.average]])
        self.textmean.SetLabel(str(self.average))
        self.OnDraw()

    def UpdateDropdown(self, e=None):
        self.BGlist = list()
        # self.BGlist.append("File/User")
        for item in self.parent.Background:
            self.BGlist.append(item.name)
        self.dropdown.SetItems(self.BGlist)
        # Show the last item
        self.dropdown.SetSelection(len(self.BGlist)-1)


def ApplyAutomaticBackground(page, bg, parent):
    """
        Creates an "automatic" background with countrate in kHz *bg* and
        applies it to the given *page* object. If an automatic
        background with the same countrate exists, uses it.

        Input:
        *page*   - page to which the background should be applied
        *bg*     - background that should be applied to that page
                   float or list of 1 or two elements
                   -> if the page is cross-correlation, the second
                      background will be applied as well.
        *parent* - parent containing *Background* list
    """
    bglist = 1*np.atleast_1d(bg)
    # minus 1 to identify non-set background id
    bgid = np.zeros(bglist.shape, dtype=int) - 1
    for b in range(len(bglist)):
        bgname = "AUTO: {:e} kHz \t".format(bglist[b])
        # Check if exists:
        for i in range(len(parent.Background)):
            if (parent.Background[i].countrate == bglist[b] and
                    parent.Background[i].name == bgname):
                bgid[b] = i
        if bgid[b] == -1:
            # Add new background
            parent.Background.append(
                Trace(countrate=bglist[b], name=bgname, duration=1))
            bgid[b] = len(parent.Background) - 1

    # Apply background to page
    # Last item is id of background

    page.bgselected = bgid[0]

    if len(bgid) == 2:
        page.bg2selected = bgid[1]
    else:
        page.bg2selected = None

    CleanupAutomaticBackground(parent)
    page.OnAmplitudeCheck("init")
    page.PlotAll()


def CleanupAutomaticBackground(parent):
    """
        Goes through the pagelist *parent.notebook.GetPageCount()*
        and checks *parent.Background* for unnused automatic
        backgrounds.
        Removes these and updates the references to all backgrounds
        within the pages.
    """
    # Create a dictionary with keys: indices of old background list -
    # and elements: list of pages having this background
    BGdict = dict()
    for i in range(len(parent.Background)):
        BGdict[i] = list()
    # Append pages to the lists inside the dictionary
    for i in range(parent.notebook.GetPageCount()):
        Page = parent.notebook.GetPage(i)
        if Page.bgselected is not None:
            if Page.bgselected not in BGdict:
                BGdict[Page.bgselected] = list()
            BGdict[Page.bgselected].append([Page, 1])
        if Page.bg2selected is not None:
            if Page.bg2selected not in BGdict:
                BGdict[Page.bg2selected] = list()
            BGdict[Page.bg2selected].append([Page, 2])

    oldBackground = parent.Background
    parent.Background = list()
    bgcounter = 0
    for key in BGdict.keys():
        if len(BGdict[key]) != 0 or not oldBackground[key].name.endswith("\t"):
            parent.Background.append(oldBackground[key])
            for Page, bgid in BGdict[key]:
                if bgid == 1:
                    Page.bgselected = bgcounter
                else:
                    Page.bg2selected = bgcounter
            bgcounter += 1
    # If the background correction tool is open, update the list
    # of backgrounds.
    # (self.MyName="BACKGROUND")
    toolkeys = parent.ToolsOpen.keys()
    if len(toolkeys) == 0:
        pass
    else:
        for key in toolkeys:
            tool = parent.ToolsOpen[key]
            try:
                if tool.MyName == "BACKGROUND":
                    tool.UpdateDropdown()
                    tool.OnPageChanged()
            except:
                pass



""" 
This file contains a TIR-FCS kineteics model function according to:
"Total Internal Reflection Fluorescence Correlation Spectroscopy: Effects
of Lateral Diffusion and Surface-Generated Fluorescence"
Jonas Ries, Eugene P. Petrov, and Petra Schwille
Biophysical Journal, Volume 95, July 2008, 390–399
"""
import numpy as np
import scipy.special as sps
import numpy.lib.scimath as nps


def wixi(x):
    """ Complex Error Function (Faddeeva/Voigt).
        w(i*x) = exp(x**2) * ( 1-erf(x) )
        This function is called by other functions within this module.
        We are using the scipy.special.wofz module which calculates
        w(z) = exp(-z**2) * ( 1-erf(-iz) )
        z = i*x
    """
    z = x*1j
    wixi = sps.wofz(z)
    # We should have a real solution. Make sure nobody complains about
    # some zero-value imaginary numbers.
    return np.real_if_close(wixi)


# Lateral correlation function
def CF_gxy_square(parms, tau):
    u""" 2D free diffusion measured with a square pinhole.
        For the square pinhole, the correlation function can readily be
        calculated for a TIR-FCS setup.
        This function is called by other functions within this module.
        Attention:
        This is NOT g2D (or gCC), the non normalized correlation function.
        g2D = gxy * eta**2 * Conc,
        where eta is the molecular brightness, Conc the concentration and
        gxy is this function.

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] D      Diffusion coefficient
        [1] sigma  lateral size of the point spread function
                   sigma = simga_0 * lambda / NA
        [2] a      side size of the square pinhole
        *tau* - lag time

        Returns: Nonnormalized Lateral correlation function w/square pinhole.
    """
    D = parms[0]
    sigma = parms[1]
    a = parms[2]

    var1 = sigma**2+D*tau
    AA = 2*np.sqrt(var1)/(a**2*np.sqrt(np.pi))
    BB = np.exp(-a**2/(4*(var1))) - 1
    CC = sps.erf(a/(2*np.sqrt(var1)))/a
    # gx = AA*BB+CC
    # gxy = gx**2
    return (AA*BB+CC)**2


def CF_gz_CC(parms, tau, wixi=wixi):
    u""" Axial (1D) diffusion in a TIR-FCS setup.
        From Two species (bound/unbound) this is the bound part.
        This function is called by other functions within this module.

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] D_3D     3D Diffusion coefficient (species A)
        [1] D_2D     2D Diffusion coefficient of bound species C
        [2] sigma    lateral size of the point spread function
                     sigma = simga_0 * lambda / NA
        [3] a        side size of the square pinhole
        [4] d_eva    evanescent decay length (decay to 1/e)
        [5] Conc_3D  3-dimensional concentration of species A
        [6] Conc_2D  2-dimensional concentration of species C
        [7] eta_3D   molecular brightness of species A
        [8] eta_2D   molecular brightness of species C
        [9] k_a      Surface association rate constant
        [10] k_d     Surface dissociation rate constant
        *tau* - lag time
    """
    D = parms[0]
    # D_2D = parms[1]
    # sigma = parms[2]
    # a = parms[3]
    # d_eva = parms[4]
    Conc_3D = parms[5]      # ligand concentration in solution
    Conc_2D = parms[6]
    eta_3D = parms[7]
    eta_2D = parms[8]
    k_a = parms[9]
    k_d = parms[10]
    # Define some other constants:
    K = k_a/k_d              # equilibrium constant
    Beta = 1/(1 + K*Conc_3D)  # This is wrong in the Ries paper
    # Re = D / d_eva**2
    Rt = D * (Conc_3D / (Beta * Conc_2D))**2
    Rr = k_a * Conc_3D + k_d
    # Define even more constants:
    sqrtR1 = -Rr/(2*nps.sqrt(Rt)) + nps.sqrt(Rr**2/(4*Rt) - Rr)
    sqrtR2 = -Rr/(2*nps.sqrt(Rt)) - nps.sqrt(Rr**2/(4*Rt) - Rr)
    R1 = sqrtR1 ** 2
    R2 = sqrtR2 ** 2
    # Calculate return function
    A1 = eta_2D * Conc_2D / (eta_3D * Conc_3D) * Beta
    A2 = sqrtR1 * wixi(-nps.sqrt(tau*R2)) - sqrtR2 * wixi(-nps.sqrt(tau*R1))
    A3 = sqrtR1 - sqrtR2
    Sol = A1 * A2 / A3
    # There are some below numerical errors-imaginary numbers.
    # We do not want them.
    return np.real_if_close(Sol)


def CF_gz_AC(parms, tau, wixi=wixi):
    u""" Axial (1D) diffusion in a TIR-FCS setup.
        From Two species (bound/unbound) this is the cross correlation part.
        This function is called by other functions within this module.

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] D_3D     3D Diffusion coefficient (species A)
        [1] D_2D     2D Diffusion coefficient of bound species C
        [2] sigma    lateral size of the point spread function
                     sigma = simga_0 * lambda / NA
        [3] a        side size of the square pinhole
        [4] d_eva    evanescent decay length (decay to 1/e)
        [5] Conc_3D  3-dimensional concentration of species A
        [6] Conc_2D  2-dimensional concentration of species C
        [7] eta_3D   molecular brightness of species A
        [8] eta_2D   molecular brightness of species C
        [9] k_a      Surface association rate constant
        [10] k_d     Surface dissociation rate constant
        *tau* - lag time
    """
    D = parms[0]
    # D_2D = parms[1]
    # sigma = parms[2]
    # a = parms[3]
    d_eva = parms[4]
    Conc_3D = parms[5]      # ligand concentration in solution
    Conc_2D = parms[6]
    eta_3D = parms[7]
    eta_2D = parms[8]
    k_a = parms[9]
    k_d = parms[10]
    # Define some other constants:
    K = k_a/k_d             # equilibrium constant
    Beta = 1/(1 + K*Conc_3D)
    Re = D / d_eva**2
    Rt = D * (Conc_3D / (Beta * Conc_2D))**2
    Rr = k_a * Conc_3D + k_d
    # Define even more constants:
    sqrtR1 = -Rr/(2*nps.sqrt(Rt)) + nps.sqrt(Rr**2/(4*Rt) - Rr)
    sqrtR2 = -Rr/(2*nps.sqrt(Rt)) - nps.sqrt(Rr**2/(4*Rt) - Rr)
    R1 = sqrtR1 ** 2
    R2 = sqrtR2 ** 2
    # And even more more:
    sqrtR3 = sqrtR1 + nps.sqrt(Re)
    sqrtR4 = sqrtR2 + nps.sqrt(Re)
    # R3 = sqrtR3 **2
    # R4 = sqrtR4 **2
    # Calculate return function
    A1 = eta_2D * Conc_2D * k_d / (eta_3D * Conc_3D)
    A2 = sqrtR4*wixi(-nps.sqrt(tau*R1)) - sqrtR3*wixi(-nps.sqrt(tau*R2))
    A3 = (sqrtR1 - sqrtR2) * wixi(nps.sqrt(tau*Re))
    A4 = (sqrtR1 - sqrtR2) * sqrtR3 * sqrtR4
    Solution = A1 * (A2 + A3) / A4
    # There are some below numerical errors-imaginary numbers.
    # We do not want them.
    return np.real_if_close(Solution)


def CF_gz_AA(parms, tau, wixi=wixi):
    u""" Axial (1D) diffusion in a TIR-FCS setup.
        From Two species (bound/unbound) this is the unbound part.
        This function is called by other functions within this module.

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] D_3D     3D Diffusion coefficient (species A)
        [1] D_2D     2D Diffusion coefficient of bound species C
        [2] sigma    lateral size of the point spread function
                     sigma = simga_0 * lambda / NA
        [3] a        side size of the square pinhole
        [4] d_eva    evanescent decay length (decay to 1/e)
        [5] Conc_3D  3-dimensional concentration of species A
        [6] Conc_2D  2-dimensional concentration of species C
        [7] eta_3D   molecular brightness of species A
        [8] eta_2D   molecular brightness of species C
        [9] k_a      Surface association rate constant
        [10] k_d     Surface dissociation rate constant
        *tau* - lag time
    """
    D = parms[0]
    # D_2D = parms[1]
    # sigma = parms[2]
    # a = parms[3]
    d_eva = parms[4]
    Conc_3D = parms[5]      # ligand concentration in solution
    Conc_2D = parms[6]
    eta_3D = parms[7]
    eta_2D = parms[8]
    k_a = parms[9]
    k_d = parms[10]
    # Define some other constants:
    d = d_eva
    K = k_a/k_d             # equilibrium constant
    Beta = 1/(1 + K*Conc_3D)
    Re = D / d_eva**2
    Rt = D * (Conc_3D / (Beta * Conc_2D))**2
    Rr = k_a * Conc_3D + k_d
    # Define even more constants:
    sqrtR1 = -Rr/(2*nps.sqrt(Rt)) + nps.sqrt(Rr**2/(4*Rt) - Rr)
    sqrtR2 = -Rr/(2*nps.sqrt(Rt)) - nps.sqrt(Rr**2/(4*Rt) - Rr)
    R1 = sqrtR1 ** 2
    R2 = sqrtR2 ** 2
    # And even more more:
    sqrtR3 = sqrtR1 + nps.sqrt(Re)
    sqrtR4 = sqrtR2 + nps.sqrt(Re)
    R3 = sqrtR3 ** 2
    R4 = sqrtR4 ** 2
    # Calculate return function
    Sum1 = d * nps.sqrt(Re*tau/np.pi)
    Sum2 = -d/2*(2*tau*Re - 1) * wixi(np.sqrt(tau*Re))
    Sum3Mult1 = - eta_2D * Conc_2D * k_d / (eta_3D * Conc_3D *
                                            (sqrtR1 - sqrtR2))
    S3M2S1M1 = sqrtR1/R3
    S3M2S1M2S1 = wixi(-nps.sqrt(tau*R1)) + -2*nps.sqrt(tau*R3/np.pi)
    S3M2S1M2S2 = (2*tau*sqrtR1*nps.sqrt(Re) + 2*tau*Re - 1) * \
        wixi(nps.sqrt(tau*Re))
    S3M2S2M1 = -sqrtR2/R4
    S3M2S2M2S1 = wixi(-nps.sqrt(tau*R2)) + -2*nps.sqrt(tau*R4/np.pi)
    S3M2S2M2S2 = (2*tau*sqrtR2*nps.sqrt(Re) + 2*tau*Re - 1) * \
        wixi(nps.sqrt(tau*Re))
    Sum3 = Sum3Mult1 * (S3M2S1M1 * (S3M2S1M2S1 + S3M2S1M2S2) +
                        S3M2S2M1 * (S3M2S2M2S1 + S3M2S2M2S2))
    Sum = Sum1 + Sum2 + Sum3
    # There are some below numerical errors-imaginary numbers.
    # We do not want them.
    return np.real_if_close(Sum)


# 3D-2D binding/unbinding TIRF
def CF_Gxyz_TIR_square_ubibi(parms, tau,
                             gAAz=CF_gz_AA, gACz=CF_gz_AC, gCCz=CF_gz_CC,
                             gxy=CF_gxy_square):
    u""" Two-component two- and three-dimensional diffusion
        with a square-shaped lateral detection area taking into account
        the size of the point spread function; and an exponential
        decaying profile in axial direction. This model covers binding
        and unbinding kintetics. 

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] D_3D    Diffusion coefficient of freely diffusing species A
        [1] D_2D    Diffusion coefficient of bound species C
        [2] σ       Lateral size of the point spread function
                    σ = σ₀ * λ / NA
        [3] a       Side size of the square-shaped detection area 
        [4] d_eva   Evanescent decay constant
        [5] C_3D    Concentration of species A in observation volume
        [6] C_2D    Concentration of species C in observation area
        [7] η_3D    Molecular brightness of species A
        [8] η_2D    Molecular brightness of species C
        [9] k_a     Surface association rate constant
        [10] k_d    Surface dissociation rate constant
        *tau* - lag time

        Returns: 3D correlation function for TIR-FCS w/square pinhole and
                 surface binding/unbinding.

        Model introduced in:
         Jonas Ries, Eugene P. Petrov, and Petra Schwille
         Total Internal Reflection Fluorescence Correlation Spectroscopy: 
         Effects of Lateral Diffusion and Surface-Generated Fluorescence
         Biophysical Journal, Vol.95, July 2008, 390–399
    """
    D_3D = parms[0]
    D_2D = parms[1]
    sigma = parms[2]
    a = parms[3]
    kappa = 1/parms[4]
    Conc_3D = parms[5]
    Conc_2D = parms[6]
    eta_3D = parms[7]
    eta_2D = parms[8]
    # k_a = parms[9]
    # k_d = parms[10]
    # We now need to copmute a real beast:
    # Inter species non-normalized correlation functions
    # gAA = gAAz * gxy(D_3D)
    # gAC = gACz * np.sqrt ( gxy(D_3D) * gxy(D_2D) )
    # gCC = gCCz * gxy(D_2D)
    # Nonnormalized correlation function
    # g = eta_3D * Conc_3D * ( gAA + 2*gAC + gCC )
    # Expectation value of fluorescence signal
    # F = eta_3D * Conc_3D / kappa + eta_2D * Conc_2D
    # Normalized correlation function
    # G = g / F**2
    ##
    # Inter species non-normalized correlation functions
    # The gijz functions take the same parameters as this function
    # The gxy function needs two different sets of parameters, depending
    # on the diffusion constant used.
    #    [0] D: Diffusion coefficient
    #    [1] sigma: lateral size of the point spread function
    #    [3] a: side size of the square pinhole
    parms_xy_2D = [D_2D, sigma, a]
    parms_xy_3D = [D_3D, sigma, a]
    # Here we go.
    gAA = gAAz(parms, tau) * gxy(parms_xy_3D, tau)
    gAC = gACz(parms, tau) * nps.sqrt(gxy(parms_xy_3D, tau) *
                                      gxy(parms_xy_2D, tau))
    gCC = gCCz(parms, tau) * gxy(parms_xy_2D, tau)
    # Nonnormalized correlation function
    g = eta_3D * Conc_3D * (gAA + 2*gAC + gCC)
    # Expectation value of fluorescence signal
    F = eta_3D * Conc_3D / kappa + eta_2D * Conc_2D
    # Normalized correlation function
    G = g / F**2
    # There are some below numerical errors-imaginary numbers.
    # We do not want them.
    return G.real
    # FNEW = eta_3D * Conc_3D / kappa
    # GNEW = eta_3D * Conc_3D * gCCz(parms, tau) / FNEW**2
    # return GNEW.real


# 3D-2D binding Model TIR
m_tir_3d_2d_ubib6021 = [6021, u"3D+2D+kin",
                        "Surface binding and unbinding, 3D TIR",
                        CF_Gxyz_TIR_square_ubibi]
labels_6021 = [u"D_3D [10 µm²/s]",
               u"D_2D [10 µm²/s]",
               u"σ [100 nm]",
               u"a [100 nm]",
               u"d_eva [100 nm]",
               u"C_3D [1000 /µm³]",
               u"C_2D[100 /µm²]",
               u"η_3D",
               u"η_2D",
               u"k_a [µm³/s]",
               u"k_d [10³ /s]"
               ]
values_6021 = [
    9.0,      # D_3D [10 µm²/s]
    0.0,      # D_2D [10 µm²/s]
    2.3,      # σ [100 nm]
    7.50,     # a [100 nm]
    1.0,      # d_eva [100 nm]
    0.01,     # conc.3D [1000 /µm³]
    0.03,     # conc.2D [100 /µm²]
    1,        # η_3D
    1,        # η_2D
    0.00001,  # k_a [µm³/s]
    0.000064  # k_d [10³ /s]
]
valuestofit_6021 = [False, True, False, False, False, False, True, False,
                    False, False, False]
# For user comfort we add values that are human readable.
# Theese will be used for output that only humans can read.
labels_human_readable_6021 = [
    u"D_3D [µm²/s]",
    u"D_2D [µm²/s]",
    u"σ [nm]",
    u"a [nm]",
    u"d_eva [nm]",
    u"C_3D [1/µm³]",
    u"C_2D [1/µm²]",
    u"η_3D",
    u"η_2D",
    u"k_a [µm³/s]",
    u"k_d [1/s]"
]
values_factor_human_readable_6021 = [10,  # "D_3D [µm²/s]",
                                     10,     # D_2D [10 µm²/s]
                                     100,    # σ [100 nm]
                                     100,    # a [100 nm]
                                     100,    # d_eva [100 nm]
                                     1000,   # conc.3D [1000 /µm³]
                                     100,    # conc.2D [100 /µm²]
                                     1,      # η_3D
                                     1,      # η_2D
                                     1,      # k_a [µm³/s]
                                     1000   # k_d [10³ /s]
                                     ]
parms_6021 = [labels_6021, values_6021, valuestofit_6021,
              labels_human_readable_6021, values_factor_human_readable_6021]

boundaries = [[0, np.inf]] * len(values_6021)

model1 = dict()
model1["Parameters"] = parms_6021
model1["Definitions"] = m_tir_3d_2d_ubib6021
model1["Boundaries"] = boundaries
model1["Constraints"] = [[1, "<", 0]]

Modelarray = [model1]



"""Module tools - batch: batch processing"""


import numpy as np
import os
import wx

from pycorrfit import openfile as opf
from pycorrfit import models as mdls
from pycorrfit import Fit
from pycorrfit.gui.threaded_progress import ThreadedProgressDlg

# Menu entry name
MENUINFO = ["B&atch control", "Batch fitting."]


class BatchCtrl(wx.Frame):
    def __init__(self, parent):
        # Parent is main frame
        self.parent = parent
        # Get the window positioning correctly
        pos = self.parent.GetPosition()
        pos = (pos[0]+100, pos[1]+100)
        wx.Frame.__init__(self, parent=parent, title="Batch control",
                          pos=pos, style=wx.DEFAULT_FRAME_STYLE | wx.FRAME_FLOAT_ON_PARENT)
        # MYID
        # This ID is given by the parent for an instance of this class
        self.MyID = None
        # Misc
        try:
            self.curpage = self.parent.notebook.GetCurrentPage()
        except:
            self.curpage = None
        # Controls
        panel = wx.Panel(self)
        self.panel = panel
        # Icon
        self.Redraw()
        if parent.MainIcon is not None:
            wx.Frame.SetIcon(self, parent.MainIcon)
        self.Show(True)

    def GetParameters(self):
        """ The parameters
        """
        # Get the item from the dropdown list
        item = self.dropdown.GetSelection()
        if self.rbtnhere.Value == True:
            # Get parameters from this session
            if item <= 0:
                Page = self.parent.notebook.GetCurrentPage()
            else:
                Page = self.parent.notebook.GetPage(item-1)
            # First apply the parameters of the page
            Page.apply_parameters()
            # Get all parameters
            Parms = self.parent.PackParameters(Page)
        else:
            # Get Parameters from different session
            Parms = self.YamlParms[item]
        return Parms

    def GetProtectedParameterIDs(self):
        """ The model parameters that are protected from batch control

        """
        pbool = [not cb.GetValue() for cb in self.wxParameterCheckBoxes]
        return np.array(pbool, dtype=bool)

    def OnApply(self, event):
        wx.BeginBusyCursor()
        Parms = self.GetParameters()
        modelid = Parms[1]
        # Set all parameters for all pages
        for i in np.arange(self.parent.notebook.GetPageCount()):
            OtherPage = self.parent.notebook.GetPage(i)
            if (OtherPage.corr.fit_model.id == modelid and
                    OtherPage.corr.correlation is not None):
                # create a copy of the fitting parameters in
                # case we want to protect them
                proparms = OtherPage.corr.fit_parameters
                self.parent.UnpackParameters(Parms, OtherPage)
                if OtherPage.prevent_batch_modification:
                    # write back protected parameters
                    OtherPage.corr.fit_parameters = proparms
                else:
                    # write back only selected parameters
                    pbool = self.GetProtectedParameterIDs()
                    OtherPage.corr.fit_parameters[pbool] = proparms[pbool]
                OtherPage.apply_parameters_reverse()
                OtherPage.PlotAll(trigger="parm_batch")
        # Update all other tools fit the finalize trigger.
        self.parent.OnFNBPageChanged(trigger="parm_finalize")
        wx.EndBusyCursor()

    def OnClose(self, event=None):
        self.parent.toolmenu.Check(self.MyID, False)
        self.parent.ToolsOpen.__delitem__(self.MyID)
        self.Destroy()

    def OnFit(self, event):
        item = self.dropdown.GetSelection()
        if self.rbtnhere.Value == True:
            if item <= 0:
                page = self.parent.notebook.GetCurrentPage()
            else:
                page = self.parent.notebook.GetPage(item-1)
            # Get internal ID
            modelid = page.corr.fit_model.id
        else:
            # Get external ID
            modelid = self.YamlParms[item][1]

        # Get all pages with right modelid
        fit_page_list = []
        for ii in np.arange(self.parent.notebook.GetPageCount()):
            pageii = self.parent.notebook.GetPage(ii)
            if (pageii.corr.fit_model.id == modelid and
                    pageii.corr.correlation is not None):
                fit_page_list.append(pageii)

        FitProgressDlg(self, fit_page_list, trigger="fit_batch")

        if self.parent.MenuAutocloseTools.IsChecked():
            # Autoclose
            self.OnClose()

    def OnPageChanged(self, Page=None, trigger=None):
        """
            This function is called, when something in the panel
            changes. The variable `trigger` is used to prevent this
            function from being executed to save stall time of the user.
            Forr a list of possible triggers, see the doc string of
            `tools`.
        """
        if self.parent.notebook.GetPageCount() == 0:
            self.panel.Disable()
            # nothing to do
            return
        else:
            self.panel.Enable()
        # Filter triggers
        if trigger in ["fit_batch", "fit_finalize", "init",
                       "parm_batch", "parm_finalize"]:
            return

        oldpage = self.curpage
        self.curpage = self.parent.notebook.GetCurrentPage()

        # redraw this tool if necessary
        if oldpage is not None and oldpage:
            oldmodelid = oldpage.modelid
        else:
            oldmodelid = 0
        newmodelid = self.curpage.modelid
        if oldmodelid != newmodelid:
            self.RedrawParameterBox()

        # We need to update the list of Pages in self.dropdown
        if self.rbtnhere.Value == True:
            DDlist = list()
            DDlist.append("Current page")
            for i in np.arange(self.parent.notebook.GetPageCount()):
                aPage = self.parent.notebook.GetPage(i)
                DDlist.append(aPage.counter+aPage.tabtitle.GetValue())
            self.dropdown.SetItems(DDlist)
            self.dropdown.SetSelection(0)

    def OnRadioHere(self, event=None):
        self.OnPageChanged(trigger="view")
        self.RedrawParameterBox()

    def OnRadioThere(self, event=None):
        # If user clicks on pages in main program, we do not want the list
        # to be changed.
        wc = opf.session_wildcards
        wcstring = "PyCorrFit session (*.pcfs)|*{};*{}".format(
            wc[0], wc[1])
        dlg = wx.FileDialog(self.parent, "Open session file",
                            self.parent.dirname, "", wcstring, wx.FD_OPEN)
        # user cannot do anything until he clicks "OK"
        if dlg.ShowModal() == wx.ID_OK:
            sessionfile = dlg.GetPath()
            self.dirname = os.path.split(sessionfile)[0]
        else:
            self.parent.dirname = dlg.GetDirectory()
            self.rbtnhere.SetValue(True)
            return

        Infodict = opf.LoadSessionData(sessionfile,
                                       parameters_only=True)
        self.YamlParms = Infodict["Parameters"]
        DDlist = list()
        for i in np.arange(len(self.YamlParms)):
            # Rebuild the list
            modelid = self.YamlParms[i][1]
            modelname = mdls.modeldict[modelid][1]
            DDlist.append(self.YamlParms[i][0]+modelname)
        self.dropdown.SetItems(DDlist)
        # Set selection text to first item
        self.dropdown.SetSelection(0)
        self.RedrawParameterBox()

    def Redraw(self, e=None):
        panel = self.panel
        for child in panel.GetChildren():
            panel.RemoveChild(child)
            child.Destroy()

        # Parameter source selection
        boxleft = wx.StaticBox(panel, label="Parameter source")
        self.rbtnhere = wx.RadioButton(panel, -1, 'This session',
                                       style=wx.RB_GROUP)
        self.rbtnhere.SetValue(True)
        self.rbtnthere = wx.RadioButton(panel, -1, 'Other session')
        self.dropdown = wx.ComboBox(panel, -1, "Current page", (15, 30),
                                    wx.DefaultSize, [], wx.CB_DROPDOWN | wx.CB_READONLY)
        # Create the dropdownlist
        text2 = wx.StaticText(panel, label="""Only data sets that have the
same model as the parameter
source will be affected by
batch modification, which
includes parameter values as
well as settings for fitting
and background correction.
To prevent batch modification
of parameter values for an
individual page, check its
"prevent batch modification"
check box.""")
        self.Bind(wx.EVT_RADIOBUTTON, self.OnRadioHere, self.rbtnhere)
        self.Bind(wx.EVT_RADIOBUTTON, self.OnRadioThere, self.rbtnthere)
        self.Bind(wx.EVT_COMBOBOX, self.RedrawParameterBox, self.dropdown)
        leftSizer = wx.StaticBoxSizer(boxleft, wx.VERTICAL)
        leftSizer.Add(self.rbtnhere)
        leftSizer.Add(self.rbtnthere)
        leftSizer.AddSpacer(5)
        leftSizer.Add(self.dropdown)
        leftSizer.AddSpacer(5)
        leftSizer.Add(text2)
        leftSizer.AddSpacer(5)

        # Parameter selection
        boxright = wx.StaticBox(panel, label="Selected parameters")
        rightSizer = wx.StaticBoxSizer(boxright, wx.VERTICAL)
        self.parameter_sizer = rightSizer
        self.RedrawParameterBox()

        # Buttons
        btnapply = wx.Button(panel, wx.ID_ANY, 'Apply to applicable pages')
        btnfit = wx.Button(panel, wx.ID_ANY, 'Fit applicable pages')
        # Bindings
        self.Bind(wx.EVT_BUTTON, self.OnApply, btnapply)
        self.Bind(wx.EVT_BUTTON, self.OnFit, btnfit)

        # Sizers
        sizer_bag = wx.GridBagSizer(hgap=5, vgap=5)
        sizer_bag.Add(leftSizer, (0, 0))
        sizer_bag.Add(rightSizer, (0, 1))
        horsizer = wx.BoxSizer(wx.HORIZONTAL)
        horsizer.Add(btnapply)
        horsizer.Add(btnfit)
        sizer_bag.Add(horsizer, (1, 0), span=wx.GBSpan(1, 2))

        panel.SetSizer(sizer_bag)
        sizer_bag.Fit(panel)
        self.SetMinSize(sizer_bag.GetMinSize())
        # Check if we even have pages.
        self.OnPageChanged()
        panel.Layout()
        sizer_bag.Fit(self)
        self.mastersizer = sizer_bag
        self.mastersizer.Fit(self)

    def RedrawParameterBox(self, e=None):
        sizer = self.parameter_sizer
        panel = self.panel
        for child in sizer.GetChildren():
            window = child.GetWindow()
            panel.RemoveChild(window)
            sizer.RemoveWindow(window)
            window.Destroy()

        text = wx.StaticText(panel, label="""If desired, (de)select
individual parameters
for batch modification.""")
        sizer.Add(text)

        if self.parent.notebook.GetPageCount():
            # Get parameters of current page
            parms = self.GetParameters()
            modelid = parms[1]
            ptext, _pval = mdls.GetHumanReadableParms(modelid, parms[2])
            ptext = [p.split()[0] for p in ptext]
            self.wxParameterCheckBoxes = []
            for p in ptext:
                cb = wx.CheckBox(panel, label=p)
                cb.SetValue(True)
                self.wxParameterCheckBoxes.append(cb)
                sizer.Add(cb)

        # Try to set sizes correctly
        box = sizer.GetStaticBox()
        boxs = box.GetBestSize()
        sizs = sizer.GetMinSize()
        thesize = (max(boxs[0], sizs[0]), sizs[1])
        sizer.SetMinSize(thesize)
        box.SetSize(thesize)

        try:
            self.mastersizer.Fit(panel)
            panel.Layout()
            self.SetSize(panel.GetSize())
            self.mastersizer.Fit(self)
        except:
            pass


class FitProgressDlg(ThreadedProgressDlg):
    def __init__(self, parent, pages, trigger=None):
        """ A progress dialog for fitting in PyCorrFit

        This is a convenience class that wraps around `ThreadedProgressDlg`
        and performs all necessary steps for fitting single pages in PyCorrFit.

        Parameters
        ----------
        parent : wx object
            The parent of the progress dialog.
        pages : list of instances of `pycorrfit.gui.page.FittingPanel`
            The pages with the model and correlation for fitting.
        trigger : str
            PyCorrFit internal trigger string.
        """
        if not isinstance(pages, list):
            pages = [pages]
        self.pages = pages
        self.trigger = trigger
        title = "Fitting data"
        messages = ["Fitting page #{}.".format(
            pi.counter.strip("# :")) for pi in pages]
        targets = [Fit]*len(pages)
        args = [pi.corr for pi in pages]
        # write parameters from page instance to correlation
        [pi.apply_parameters() for pi in self.pages]
        super(FitProgressDlg, self).__init__(parent, targets, args,
                                             title=title,
                                             messages=messages)

    def finalize(self):
        """ Do everything that is required after fitting, including
        cleanup of non-fitted pages.
        """
        if self.aborted:
            # we need to cleanup
            fin_index = max(0, self.index_aborted-1)
            pab = self.pages[self.index_aborted]
            pab.fit_results = None
            pab.apply_parameters()
        else:
            fin_index = len(self.pages)

        # finalize fitting
        [pi.Fit_finalize(trigger=self.trigger)
         for pi in self.pages[:fin_index]]



"""
Go through each model, vary one parameter and fit it back to the
default value of that model.
"""
import numpy as np
import pycorrfit
from pycorrfit.correlation import Correlation
from pycorrfit.fit import Fit
import pytest

# GLOBAL PARAMETERS FOR THIS TEST:
TAUMIN = 1e-3
TAUMAX = 1e6
TAULEN = 100
FITALG = "Lev-Mar"


def fit_single_parameter(modelid, fullparms, parmid, parmval, noise=False):
    """
    Use the full parameter set `fullparms` and leave a single parameter
    `parmid` variable during the fit.
    Returns the fitted value of the parameter with index `parmid`
    """
    corr = Correlation(fit_model=modelid, fit_algorithm=FITALG, verbose=0)
    tau = np.exp(np.linspace(np.log(TAUMIN), np.log(TAUMAX), TAULEN))
    # Create artificial data by using the current fit_model
    data = corr.fit_model(fullparms, tau)
    if noise:
        if noise is True:
            deltanoise = (np.max(data)-np.min(data))/20
        else:
            deltanoise = (np.max(data)-np.min(data))*noise
        anoise = (np.random.random(data.shape[0])-.5)*deltanoise
        data += anoise
    # Add artificial data to data set
    corr.correlation = np.dstack((tau, data))[0]
    # Set variable parameters
    fit_bool = np.zeros(fullparms.shape[0])
    fit_bool[parmid] = True
    corr.fit_parameters_variable = fit_bool
    fullparms_edit = fullparms.copy()
    fullparms_edit[parmid] = parmval
    corr.fit_parameters = fullparms_edit
    Fit(corr)
    return corr.fit_parameters[parmid]


def deviate_parameter(model, parmid):
    """
    Returns a deviated version of the parameter with id `parmid`.
    Performs model checks to ensure the new value is physical.
    """
    val = model.default_values[parmid]
    if val == 0:
        val += .1
    else:
        val *= .9
    return val


@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_fit_single_parameter():
    """
    Deviate a single parameter and fit it back.
    """
    allow_fail = [
        [6082, "SP"],
    ]
    faillist = list()
    for model in pycorrfit.models.models:
        fullparms = model.default_values
        for ii, val in enumerate(fullparms):
            newval = deviate_parameter(model, ii)
            fitval = fit_single_parameter(model.id, fullparms, ii,
                                          newval, noise=False)
            # print(val-fitval)
            if not np.allclose([val], [fitval]):
                if not [model.id, model.parameters[0][ii]] in allow_fail:
                    faillist.append([model.id, model.parameters[0][ii],
                                     val, fitval])
    if faillist:
        raise ValueError("Model tests failed for:\n", faillist)


def fit_single_parameter_with_noise(noise=0.005):
    """
    Deviate a single parameter and fit it back.
    """
    faillist = list()
    succlist = list()
    for model in pycorrfit.models.models:
        fullparms = model.default_values
        for ii, val in enumerate(fullparms):
            newval = deviate_parameter(model, ii)
            fitval = fit_single_parameter(model.id, fullparms, ii, newval,
                                          noise=noise)
            if not np.allclose([val], [fitval], atol=.1, rtol=.1):
                faillist.append([model.id, model.parameters[0][ii],
                                 val, fitval])
            else:
                succlist.append([model.id, model.parameters[0][ii],
                                 val, fitval])
    return succlist, faillist


@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_fit_single_parameter_with_noise_one_permille():
    succlist, faillist = fit_single_parameter_with_noise(noise=0.001)
    if len(faillist)/len(succlist) > .01:
        raise ValueError("Model tests failed for:\n", faillist)


@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_fit_single_parameter_with_noise_two_percent():
    succlist, faillist = fit_single_parameter_with_noise(noise=0.02)
    if len(faillist)/len(succlist) > .05:
        raise ValueError("Model tests failed for:\n", faillist)


@pytest.mark.filterwarnings('ignore::pycorrfit.fit.StuckParameterWarning')
def test_fit_single_parameter_with_noise_five_percent():
    succlist, faillist = fit_single_parameter_with_noise(noise=0.05)
    if len(faillist)/len(succlist) > .10:
        raise ValueError("Model tests failed for:\n", faillist)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



""" pycorrfit.models.control

Controls which fitting models are imported an in which order.
"""
import numpy as np

from .classes import Model


def append_model(modelarray):
    """ Append a new model from a modelarray. *Modelarray* has to be a list
        whose elements have two items:
        [0] parameters
        [1] some info about the model
        See separate models for more information
    """
    global values
    global valuedict
    global models
    global modeldict
    global supplement
    global boundaries
    global modeltypes

    if not isinstance(modelarray, list):
        modelarray = [modelarray]

    for datadict in modelarray:
        # We can have many models in one model array
        amod = Model(datadict)

        models.append(amod)
        if amod.id in modeldict:
            raise ValueError("Model with same is already exists: \n {} vs. {}".
                             format(amod, modeldict[amod.id]))
        modeldict[amod.id] = amod

        values.append(amod.parameters)
        valuedict[amod.id] = amod.parameters

        # Supplementary Data might be there
        supplement[amod.id] = amod.func_supplements

        # Check functions - check for correct values
        boundaries[amod.id] = amod.boundaries

        # Add model type to internal type list.
        if amod.type is not None:
            if not amod.type in modeltypes:
                modeltypes[amod.type] = []
            modeltypes[amod.type].append(amod.id)


def model_setup(modelid, name, comp, mtype, fctn, par_labels, par_values,
                par_vary=None, par_boundaries=None, par_constraints=None,
                par_hr_labels=None, par_hr_factors=None,
                supplementary_method=None,
                ):
    u"""
    This helper method does everything that is required to make a model
    available for PyCorrFit. The idea is that this method can be called from
    anywhere and thus we do not need to do the tedious work of adding models
    in the __init__.py file.

    Parameters
    ----------
    modelid : int
        Model identifier.
    name : str
        Name of the Model.
    comp : str
        Description of components of the model, e.g. "T+3D+2D"
    mtype : str
        Type of model, e.g. "Confocal (Gaussian)"
    fctn : callable
        The method that computes the model function. It must take
        two arguments. The first is of shape `par_values` and the
        second is a 2D array containing lag time and correlation.
    par_labels : list-like, strings
        The labels of each parameter in PyCorrFit dimensionless
        representation, i.e.

            unit of time        : 1 ms
            unit of inverse time: 1000 /s
            unit of distance    : 100 nm
            unit of Diff.coeff  : 10 µm²/s
            unit of inverse area: 100 /µm²
            unit of inv. volume : 1000 /µm³
    par_values : list-like, floats
        The parameter values in PyCorrFit dimensionless units.
    par_vary :  list-like, bools or None
        A list describing which parameters should be varied during
        fitting. If not given, only the first element is set to `True`.
    par_boundaries : list-like, floats
        The parameter boundaries - two values for each parameter.
        Examples: [[0, np.inf], [0,1]]
    par_constraints : list of lists
        Constraints between parameters. For example, make sure parameter
        2 is always larger than parameter 1 and parameter 5 is always
        smaller than parameter 1: [[2, ">", 1], [5, "<", 1]]
        Parameter count starts at 0.
    par_hr_labels : list-like, strings
        User-defined human readable labels of the parameters. If this is
        set, `par_hr_factors` is also required.
    par_hr_factors : list-like, floats
        The multiplicative factors to get from `par_labels` to
        `par_hr_labels`.
    supplementary_method : callable
        A method that takes the parameters `par_values` and the countrate
        of the experiment as an argument and returns a dictinoary of
        supplementary information.
    """
    # Checks
    assert len(par_labels) == len(par_values)
    for p in [par_vary,
              par_boundaries,
              par_hr_labels,
              par_hr_factors,
              ]:
        if p is not None:
            assert len(p) == len(
                par_values), "Number of parameters must match!"

    if par_hr_factors is None or par_hr_labels is None:
        assert par_hr_factors is None, "human readable requires two parameter"
        assert par_hr_labels is None, "human readable requires two parameter"

    if par_vary is None:
        # Set par_vary
        par_vary = np.zeros(len(par_values), dtype=bool)
        par_vary[0] = True

    if par_hr_factors is None:
        # Set equal to labels
        par_hr_labels = par_labels
        par_hr_factors = np.ones_like(par_values)

    model = {}

    model["Parameters"] = [par_labels, par_values, par_vary,
                           par_hr_labels, par_hr_factors]

    model["Definitions"] = [modelid, comp, name, fctn, mtype]

    if supplementary_method is not None:
        model["Supplements"] = supplementary_method

    if par_boundaries is not None:
        model["Boundaries"] = par_boundaries

    if par_constraints is not None:
        model["Constraints"] = par_constraints

    append_model(model)


# Pack all variables
values = list()
# Also create a dictionary, key is modelid
valuedict = dict()
# Pack all models
models = list()
# Also create a dictinary
modeldict = dict()
# A dictionary for supplementary data:
supplement = dict()
# A dictionary containing model boundaries
boundaries = dict()

# shorttypes are used by the GUI to abbreviate the model type
shorttype = dict()
shorttype[u"Confocal (Gaussian)"] = u"CFoc"
shorttype[u"Confocal (Gaussian) and triplet"] = u"CFoc"
shorttype[u"Confocal (Gaussian) with double triplet"] = u"CFoc"
shorttype[u"TIR (Gaussian/Exp.)"] = u"TIR CFoc"
shorttype[u"TIR (□xσ/Exp.)"] = u"TIR □xσ"

# Create a list for the differentiation between the models
# This should make everything look a little cleaner
modeltypes = {}
modeltypes[u"User"] = []

# The order of the import matters!
# These models perform the integration by themselves using the `model_setup` method.
from . import MODEL_TIRF_3D2Dkin_Ries
from . import MODEL_TIRF_3D3D
from . import MODEL_TIRF_3D2D
from . import MODEL_TIRF_2D2D
from . import MODEL_TIRF_1C
from . import MODEL_TIRF_gaussian_3D3D
from . import MODEL_TIRF_gaussian_3D2D
from . import MODEL_TIRF_gaussian_1C
from . import model_confocal_tt_3d_2d
from . import model_confocal_tt_2d_2d
from . import model_confocal_tt_2d
from . import model_confocal_tt_3d_3d
from . import model_confocal_tt_3d
from . import model_confocal_t_3d_3d_2d
from . import model_confocal_t_3d_3d_3d
from . import model_confocal_t_3d_2d
from . import model_confocal_t_2d_2d
from . import model_confocal_t_2d
from . import model_confocal_t_3d_3d
from . import model_confocal_t_3d
from . import model_confocal_3d_2d
from . import model_confocal_2d_2d
from . import model_confocal_2d
from . import model_confocal_3d_3d
from . import model_confocal_3d

# These lines can be removed once all models are converted
# from `MODEL_*` to `model_` syntax.
modeltypes[u"TIR (Gaussian/Exp.)"] = [6014, 6034, 6033]
modeltypes[u"TIR (□xσ/Exp.)"] = [6010, 6023, 6000, 6022, 6020, 6021]

# Old models

# Load all models from the imported "MODEL_*" submodules
# These are the models that were not imported using the `model_setup` method.
for g in list(globals().keys()):
    if g.startswith("MODEL_") and hasattr(globals()[g], "Modelarray"):
        append_model(globals()[g].Modelarray)



"""
PyCorrFit is a tool to fit fluorescence correlation spectroscopy
data on a logarithmic scale.
"""

from importlib.metadata import PackageNotFoundError, version

from . import meta, models, openfile, readfiles
from .correlation import Correlation
from .fit import Fit
from .trace import Trace

try:
    __version__ = version("pycorrfit")
except PackageNotFoundError:
    # package is not installed
    __version__ = "unknown"


__author__ = "Paul Müller"
__license__ = "GPL v2"
__all__ = ["meta", "models", "openfile", "readfiles", "Fit", "Trace", "Correlation"]



import numpy as np

from .control import model_setup
from .cp_confocal import threed

# 3D simple gauss


def CF_Gxyz_gauss(parms, tau):
    # Model 6012
    u""" Three-dimanesional free diffusion with a Gaussian laser profile
        (eliptical).

        G(τ) = offset + 1/( n*(1+τ/τ_diff) * sqrt(1 + τ/(SP²*τ_diff)) )

        Calculation of diffusion coefficient and concentration
        from the effective radius of the detection profile (r₀ = 2*σ):
        D = r₀²/(4*τ_diff)
        Conc = n/( sqrt(π³)*r₀²*z₀ )

            r₀   lateral detection radius (waist of lateral gaussian)
            z₀   axial detection length (waist of axial gaussian)
            D    Diffusion coefficient
            Conc Concentration of dye

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal volume
        [1] τ_diff  Characteristic residence time in confocal volume
        [2] SP      SP=z₀/r₀ Structural parameter,
                    describes the axis ratio of the confocal volume
        [3] offset
        *tau* - lag time
    """
    n = parms[0]
    taudiff = parms[1]
    SP = parms[2]
    off = parms[3]

    BB = threed(tau, taudiff, SP)

    G = off + 1/n * BB
    return G


def supplements(parms, countrate=None):
    # We can only give you the effective particle number
    n = parms[0]
    Info = list()
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [4.0, 0.4, 5.0, 0.0]
boundaries = [[0, np.inf]]*len(parms)
boundaries[-1] = [-np.inf, np.inf]

model_setup(
    modelid=6012,
    name="3D diffusion (confocal)",
    comp="3D",
    mtype="Confocal (Gaussian)",
    fctn=CF_Gxyz_gauss,
    par_labels=[
        u"n",
        u"τ_diff [ms]",
        u"SP",
        u"offset"],
    par_values=parms,
    par_vary=[True, True, False, False],
    par_boundaries=boundaries,
    supplementary_method=supplements
)



import numpy as np
import scipy.special as sps


def wixi(x):
    """ Complex Error Function (Faddeeva/Voigt).
        w(i*x) = exp(x**2) * ( 1-erf(x) )
        This function is called by other functions within this module.
        We are using the scipy.special.wofz module which calculates
        w(z) = exp(-z**2) * ( 1-erf(-iz) )
        z = i*x
    """
    z = x*1j
    wixi = sps.wofz(z)
    # We should have a real solution. Make sure nobody complains about
    # some zero-value imaginary numbers.

    return np.real_if_close(wixi)

# 3D + 2D + T
# model 6033


def CF_Gxyz_3d2dT_gauss(parms, tau):
    u""" Two-component, two- and three-dimensional diffusion
        with a Gaussian lateral detection profile and
        an exponentially decaying profile in axial direction,
        including a triplet component.
        The triplet factor takes into account blinking according to triplet
        states of excited molecules.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        kappa = 1/d_eva
        x = sqrt(D_3D*τ)*kappa
        w(i*x) = exp(x²)*erfc(x)
        gz = kappa * 
             [ sqrt(D_3D*τ/pi) + (1 - 2*D_3D*τ*kappa²)/(2*kappa) * w(i*x) ]
        g2D3D = 1 / [ 1+4*D_3D*τ/r₀² ]
        particle3D = α*F * g2D3D * gz
        particle2D = (1-F)/ (1+4*D_2D*τ/r₀²) 
        triplet = 1 + T/(1-T)*exp(-τ/τ_trip)
        norm = (1-F + α*F)²
        G = 1/n*(particle2D + particle3D)*triplet/norm + offset

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0] n       Effective number of particles in confocal volume
                    (n = n2D+n3D)
        [1] D_2D    Diffusion coefficient  of surface bound particles
        [2] D_3D    Diffusion coefficient of freely diffusing particles
        [3] F       Fraction of molecules of the freely diffusing species
                    (n3D = n*F), 0 <= F <= 1
        [4] r₀      Lateral extent of the detection volume
        [5] d_eva   Evanescent field depth
        [6] α       Relative molecular brightness of freely diffusing
                    compared to surface bound particles (α = q3D/q2D)
        [7] τ_trip  Characteristic residence time in triplet state
        [8] T       Fraction of particles in triplet (non-fluorescent) state
                    0 <= T < 1
        [9] offset
        *tau* - lag time
    """
    n = parms[0]
    D2D = parms[1]
    D3D = parms[2]
    F = parms[3]
    r0 = parms[4]
    deva = parms[5]
    alpha = parms[6]
    tautrip = parms[7]
    T = parms[8]
    off = parms[9]

    # 2D species
    taud2D = r0**2/(4*D2D)
    particle2D = (1-F) / (1+tau/taud2D)

    # 3D species
    taud3D = r0**2/(4*D3D)
    # 2D gauss component
    g2D3D = 1 / ((1.+tau/taud3D))
    # 1d TIR component
    # Axial correlation
    kappa = 1/deva
    x = np.sqrt(D3D*tau)*kappa
    w_ix = wixi(x)
    # Gz = 1/N1D * gz = kappa / Conc.1D * gz
    gz = kappa * (np.sqrt(D3D*tau/np.pi) -
                  (2*D3D*tau*kappa**2 - 1)/(2*kappa) * w_ix)
    particle3D = alpha**2*F * g2D3D * gz

    # triplet
    if tautrip == 0 or T == 0:
        triplet = 1
    else:
        triplet = 1 + T/(1-T) * np.exp(-tau/tautrip)

    # Norm
    norm = (1-F + alpha*F)**2

    # Correlation function
    G = 1/n*(particle2D + particle3D)*triplet/norm
    return G + off


def get_boundaries(parms):
    # strictly positive
    boundaries = [[0, np.inf]]*len(parms)
    # F
    boundaries[3] = [0, .9999999999999]
    # T
    boundaries[8] = [0, .9999999999999]
    boundaries[-1] = [-np.inf, np.inf]
    return boundaries


def MoreInfo(parms, countrate=None):
    u"""Supplementary parameters:
        Effective number of freely diffusing particles in 3D:
        [10] n3D = n*F
        Effective number particles diffusing on 2D surface:
        [11] n2D = n*(1-F)
        Value of the correlation function at lag time zero:
        [12] G(0)
        Effective measurement volume:
        [13] V_eff [al] = π * r₀² * d_eva
        Concentration of the 2D species:
        [14] C_2D [1/µm²] = n2D / ( π * r₀² )
        Concentration of the 3D species:
        [15] C_3D [nM] = n3D/V_eff
    """
    # We can only give you the effective particle number
    n = parms[0]
    # D2D=parms[1]
    # D3D=parms[2]
    F = parms[3]
    r0 = parms[4]
    deva = parms[5]
    # alpha=parms[6]

    Info = list()
    # The enumeration of these parameters is very important for
    # plotting the normalized curve. Countrate must come out last!
    Info.append([u"n3D", n*F])
    Info.append([u"n2D", n*(1.-F)])
    # Detection area:
    Veff = np.pi * r0**2 * deva
    C3D = n*F / Veff
    C2D = n*(1-F) / (np.pi*r0**2)
    # Correlation function at tau = 0
    G_0 = CF_Gxyz_3d2dT_gauss(parms, 0)
    Info.append([u"G(0)", G_0])
    Info.append([u"V_eff [al]", Veff])
    Info.append([u"C_2D [1/µm²]", C2D * 100])
    Info.append([u"C_3D [nM]", C3D * 10000/6.0221415])

    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append([u"cpp [kHz]", cpp])
    return Info


# 3D + 3D + T model gauss
m_gauss_3d_2d_t = [6033, "T+3D+2D",
                   "Separate 3D and 2D diffusion + triplet w/ TIR",
                   CF_Gxyz_3d2dT_gauss]
labels = [u"n",
          u"D_2D [10 µm²/s]",
          u"D_3D [10 µm²/s]",
          u"F_3D",
          u"r₀ [100 nm]",
          u"d_eva [100 nm]",
          u"\u03b1"+" (q_3D/q_2D)",
          u"τ_trip [ms]",
          u"T",
          u"offset"
          ]
values = [
    25,      # n
    0.51,       # D2D
    25.1,    # D3D
    0.45,     # F3D
    9.44,       # r0
    1.0,       # deva
    1.0,     # alpha
    0.001,       # tautrip
    0.01,       # T
    0.0      # offset
]
# Human readable stuff
labelshr = [u"n",
            u"D_2D [µm²/s]",
            u"D_3D [µm²/s]",
            u"F_3D",
            u"r₀ [nm]",
            u"d_eva [nm]",
            u"\u03b1"+u" (q_3D/q_2D)",
            u"τ_trip [µs]",
            u"T",
            u"offset"
            ]
valueshr = [
    1.,      # n
    10.,       # D2D
    10.,    # D3D
    1.,     # F3D
    100.,       # r0
    100.,       # deva
    1.,     # alpha
    1000.,       # tautrip
    1.,       # T
    1.      # offset
]

valuestofit = [True, True, True, True,
               False, False, False, False, False, False]
parms = [labels, values, valuestofit, labelshr, valueshr]


model1 = dict()
model1["Parameters"] = parms
model1["Definitions"] = m_gauss_3d_2d_t
model1["Boundaries"] = get_boundaries(values)
model1["Supplements"] = MoreInfo
model1["Constraints"] = [[2, ">", 1]]

Modelarray = [model1]



import numpy as np

from .control import model_setup
from .cp_confocal import twod
from .cp_triplet import trip
from .cp_mix import double_pnum


# 2D + 2D + TT Gauß
# Model 6044
def CF_Gxy_gauss_2D2DTT(parms, tau):
    u""" Two-component two-dimensional free diffusion
        with a Gaussian laser profile, including two triplet components.
        The triplet factor takes into account a blinking term.
        Set *T* or *τ_trip* to 0, if no triplet component is wanted.

        particle1 = F₁/(1+τ/τ₁)
        particle2 = α²*(1-F₁)/(1+τ/τ₂)
        triplet1 = 1 + T₁/(1-T₁)*exp(-τ/τ_trip₁)
        triplet2 = 1 + T₂/(1-T₂)*exp(-τ/τ_trip₂)
        norm = (F₁ + α*(1-F₁))²
        G = 1/n*(particle1 + particle2)*triplet1*triplet2/norm + offset

        *parms* - a list of parameters.
        Parameters (parms[i]):
        [0]  n        Effective number of particles in confocal volume
                      (n = n₁+n₂)
        [1]  τ₁       Diffusion time of particle species 1
        [2]  τ₂       Diffusion time of particle species 2
        [3]  F₁       Fraction of molecules of species 1 (n₁ = n*F₁)
                      0 <= F₁ <= 1
        [4]  α        Relative molecular brightness of particle
                      2 compared to particle 1 (α = q₂/q₁)
        [5]  τ_trip₁  Characteristic residence time in triplet state
        [6]  T₁       Fraction of particles in triplet (non-fluorescent) state
                      0 <= T < 1
        [7]  τ_trip₂  Characteristic residence time in triplet state
        [8]  T₂       Fraction of particles in triplet (non-fluorescent) state
                      0 <= T < 1
        [9]  offset
        *tau* - lag time
    """
    n = parms[0]
    taud1 = parms[1]
    taud2 = parms[2]
    F = parms[3]
    alpha = parms[4]
    tautrip1 = parms[5]
    T1 = parms[6]
    tautrip2 = parms[7]
    T2 = parms[8]
    off = parms[9]

    g = double_pnum(n=n,
                    F1=F,
                    alpha=alpha,
                    comp1=twod,
                    comp2=twod,
                    kwargs1={"tau": tau,
                             "taudiff": taud1},
                    kwargs2={"tau": tau,
                             "taudiff": taud2},
                    )

    tr1 = trip(tau=tau, T=T1, tautrip=tautrip1)
    tr2 = trip(tau=tau, T=T2, tautrip=tautrip2)

    G = off + g * tr1 * tr2

    return G


def supplements(parms, countrate=None):
    u"""Supplementary parameters:
        [10]  n₁ = n*F₁     Particle number of species 1
        [11] n₂ = n*(1-F₁) Particle number of species 2
    """
    # We can only give you the effective particle number
    n = parms[0]
    F1 = parms[3]
    Info = list()
    # The enumeration of these parameters is very important for
    # plotting the normalized curve. Countrate must come out last!
    Info.append([u"n\u2081", n*F1])
    Info.append([u"n\u2082", n*(1.-F1)])
    if countrate is not None:
        # CPP
        cpp = countrate/n
        Info.append(["cpp [kHz]", cpp])
    return Info


parms = [
    25,      # n
    5,       # taud1
    1000,    # taud2
    0.5,     # F
    1.0,     # alpha
    0.001,   # tautrip1
    0.01,    # T1
    0.002,   # tautrip2
    0.01,    # T2
    0.0      # offset
]

# Boundaries
# strictly positive
boundaries = [[0, np.inf]]*len(parms)
# F
boundaries[3] = [0, .9999999999999]
# T
boundaries[6] = [0, .9999999999999]
boundaries[8] = [0, .9999999999999]
boundaries[-1] = [-np.inf, np.inf]


model_setup(
    modelid=6044,
    name="Separate 2D diffusion with double triplet (confocal)",
    comp="T+T+2D+2D",
    mtype="Confocal (Gaussian) with double triplet",
    fctn=CF_Gxy_gauss_2D2DTT,
    par_labels=[
        u"n",
        u"τ"+u"\u2081"+" [ms]",
        u"τ"+u"\u2082"+" [ms]",
        u"F"+u"\u2081",
        u"\u03b1"+" (q"+u"\u2082"+"/q"+u"\u2081"+")",
        u"τ_trip₁ [ms]",
        u"T₁",
        u"τ_trip₂ [ms]",
        u"T₂",
        u"offset"
    ],
    par_values=parms,
    par_vary=[True, True, True, True, False,
              False, False, False, False, False],
    par_boundaries=boundaries,
    par_constraints=[[2, ">", 1], [5, "<", 1], [7, ">", 5]],
    par_hr_labels=[
        u"n",
        u"τ₁ [ms]",
        u"τ₂ [ms]",
        u"F₁",
        u"\u03b1"+u" (q₂/q₁)",
        u"τ_trip₁ [µs]",
        u"T₁",
        u"τ_trip₂ [µs]",
        u"T₂",
        u"offset"
    ],
    par_hr_factors=[
        1.,     # n
        1.,     # taud1
        1.,     # taud2
        1.,     # F
        1.,     # alpha
        1000.,  # tautrip1 [µs]
        1.,     # T1
        1000.,  # tautrip2 [µs]
        1.,     # T2
        1.      # offset
    ],
    supplementary_method=supplements
)



import copy
import numbers
import warnings

import numpy as np


class Model(object):
    """General class for handling FCS fitting models"""

    def __init__(self, datadict):
        """datadict is an item in Modelarray"""
        self._parameters = datadict["Parameters"]
        self._definitions = datadict["Definitions"]

        if "Supplements" in list(datadict.keys()):
            self._supplements = datadict["Supplements"]
        else:
            self._supplements = lambda x, y: []

        if "Boundaries" in list(datadict.keys()):
            self._boundaries = datadict["Boundaries"]
        else:
            # dummy verification function
            self._boundaries = [[-np.inf, np.inf]]*len(self._parameters[1])

        if "Constraints" in list(datadict.keys()):
            # sort constraints such that the first value is always
            # larger than the last.
            newcc = []
            for cc in datadict["Constraints"]:
                # check for integral numbers to avoid comparison to strings
                # in e.g. confocal t_3d_3d_3d model.
                if (isinstance(cc[0], numbers.Integral) and
                    isinstance(cc[2], numbers.Integral) and
                        cc[0] < cc[2]):
                    if cc[1] == ">":
                        cc = [cc[2], "<", cc[0]]
                    elif cc[1] == "<":
                        cc = [cc[2], ">", cc[0]]
                newcc.append(cc)
            self._constraints = newcc
        else:
            self._constraints = []

    def __call__(self, parameters, tau):
        return self.function(parameters, tau)

    def __getitem__(self, key):
        """Emulate old list behavior of models"""
        return self._definitions[key]

    def __repr__(self):
        text = "Model {} - {}".format(
            self.id,
            self.description_short)
        return text

    def apply(self, parameters, tau):
        """
        Apply the model with `parameters` and lag
        times `tau`
        """
        return self.function(parameters, tau)

    @property
    def boundaries(self):
        return self._boundaries

    @property
    def constraints(self):
        """ fitting constraints """
        return copy.copy(self._constraints)

    @property
    def components(self):
        """how many components does this model have"""
        return self._definitions[1]

    @property
    def default_values(self):
        """default fitting values"""
        return np.array(self._parameters[1]).copy()

    @property
    def default_variables(self):
        """indexes default variable fitting (bool)"""
        return np.array(self._parameters[2]).copy()

    @property
    def description_long(self):
        """long description"""
        return self._definitions[3].__doc__

    @property
    def description_short(self):
        """short description"""
        return self._definitions[2]

    @property
    def function(self):
        return self._definitions[3]

    @property
    def func_supplements(self):
        return self._supplements

    @property
    def func_verification(self):
        warnings.warn(
            "`func_verification is deprecated: please do not use it!")
        return lambda x: x

    def get_supplementary_parameters(self, values, countrate=None):
        """
        Compute additional information for the model

        Parameters
        ----------
        values: list-like of same length as `self.default_values`
            parameters for the model
        countrate: float
            countrate in kHz
        """
        return self.func_supplements(values, countrate)

    def get_supplementary_values(self, values, countrate=None):
        """
        Returns only the values of
        self.get_supplementary_parameters

        Parameters
        ----------
        values: list-like of same length as `self.default_values`
            parameters for the model
        countrate: float
            count rate in Hz
        """
        out = list()
        for item in self.get_supplementary_parameters(values, countrate):
            out.append(item[1])
        return out

    @property
    def id(self):
        return self._definitions[0]

    @property
    def name(self):
        return self.description_short

    @property
    def parameters(self):
        return self._parameters

    @property
    def type(self):
        if len(self._definitions) < 5:
            return None
        else:
            return self._definitions[4]



"""Show all images in the "gallery" folder

Usage:

   .. simple_gallery::
      :dir: gallery

where "gallery" is relative to the folder containing conf.py.

Changelog:
0.1 (2019-10-24)
 - initial release
"""
import pathlib

from docutils.statemachine import ViewList
from docutils.parsers.rst import Directive, directives
from sphinx.util.nodes import nested_parse_with_titles
from docutils import nodes


class SimpleGalleryDirective(Directive):
    required_arguments = 0
    optional_arguments = 1
    final_argument_whitespace = True

    option_spec = {
        'dir': directives.unchanged,
    }

    def get_files(self):
        """This will return a list of files"""
        root = pathlib.Path(__file__).parent.parent
        gpath = root / self.options["dir"]
        files = []
        for ff in gpath.glob("*"):
            if ff.suffix in [".png", "*.jpg"]:
                files.append(ff.relative_to(root))
        return files

    def run(self):
        rst = []
        files = self.get_files()

        for ff in files:
            rst.append(".. image:: {}".format(ff))
            rst.append("    :target: _images/{}".format(ff.name))
            rst.append("    :scale: 25%")
            rst.append("    :align: left")
            rst.append("")

        vl = ViewList(rst, "fakefile.rst")
        # Create a node.
        node = nodes.section()
        node.document = self.state.document
        # Parse the rst.
        nested_parse_with_titles(self.state, vl, node)
        return node.children


def setup(app):
    app.add_directive('simple_gallery', SimpleGalleryDirective)
    return {'version': '0.1'}   # identifies the version of our extension



from os.path import dirname, exists, join, realpath
from urllib.request import urlretrieve

from setuptools import setup

# Download documentation if it was not compiled with latex
pdfdoc = join(dirname(realpath(__file__)), "doc/PyCorrFit_doc.pdf")
webdoc = "https://github.com/FCS-analysis/PyCorrFit/wiki/PyCorrFit_doc.pdf"
if not exists(pdfdoc):
    print("Downloading {} from {}".format(pdfdoc, webdoc))
    try:
        urlretrieve(webdoc, pdfdoc)
    except Exception as _e:
        print("Failed to download documentation.")

setup()



"""Main execution script"""
from looseversion import LooseVersion
import sys
import warnings

import yaml
import numpy as np
import scipy


class Fake(object):
    """ Fake module.
    """

    def __init__(self):
        self.__version__ = "0.0 unknown"
        self.version = "0.0 unknown"
        self.use = lambda x: None


# Import matplotlib a little earlier. This way some problems with saving
# dialogs that are not made by "WXAgg" are solved.
try:
    import matplotlib
except ImportError:
    matplotlib = Fake()
# We do catch warnings about performing this before matplotlib.backends stuff
# matplotlib.use('WXAgg') # Tells matplotlib to use WxWidgets
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    matplotlib.use('WXAgg')  # Tells matplotlib to use WxWidgets for dialogs


# Sympy is optional:
try:
    import sympy
except ImportError:
    print("Importing sympy failed! Checking of external model functions")
    print("will not work!")
    # We create a fake module sympy with a __version__ property.
    # This way users can run PyCorrFit without having installed sympy.
    sympy = Fake()

# We must not import wx here. frontend/gui does that. If we do import wx here,
# somehow unicode characters will not be displayed correctly on windows.
# import wx

# Continue with the import:
from . import frontend as gui  # noqa: E402
from . import doc  # noqa: E402


def CheckVersion(given, required, name):
    """ For a given set of versions  str *required* and str *given*,
    where version are usually separated by dots, print whether for
    the module str *name* the required verion is met or not.
    """
    try:
        req = LooseVersion(required)
        giv = LooseVersion(given)
    except:
        print(" WARNING: Could not verify version of "+name+".")
        return
    if req > giv:
        print(" WARNING: You are using "+name+" v. "+given +
              " | Required: "+name+" " + required)
    else:
        print(" OK: "+name+" v. "+given+" | "+required+" required")


# Start gui
def Main():

    # VERSION
    version = doc.__version__
    __version__ = version

    print(gui.doc.info(version))

    # Check important module versions
    print("\n\nChecking module versions...")
    CheckVersion(matplotlib.__version__, "2.2.2", "matplotlib")
    CheckVersion(np.__version__, "1.14.2", "NumPy")
    CheckVersion(yaml.__version__, "3.12", "PyYAML")
    CheckVersion(scipy.__version__, "1.0.1", "SciPy")
    CheckVersion(sympy.__version__, "1.1.1", "sympy")
    CheckVersion(gui.wx.__version__, "4.0.1", "wxPython")

    # Start gui
    app = gui.MyApp(False)

    frame = gui.MyFrame(None, -1, version)
    app.frame = frame

    # Before starting the main loop, check for possible session files
    # in the arguments.
    sysarg = sys.argv
    for arg in sysarg:
        if arg.endswith(".pcfs"):
            print("\nLoading Session "+arg)
            frame.OnOpenSession(sessionfile=arg)
            break
        if arg.endswith(".fcsfit-session.zip"):
            print("\nLoading Session "+arg)
            frame.OnOpenSession(sessionfile=arg)
            break
        elif arg[:6] == "python":
            pass
        elif arg[-12:] == "PyCorrFit.py":
            pass
        elif arg[-11:] == "__main__.py":
            pass
        else:
            print("Ignoring command line parameter: "+arg)

    app.MainLoop()


if __name__ == "__main__":
    Main()
